from __future__ import annotations

from types import SimpleNamespace
from typing import Any, Union

import numpy as np
from ase import Atoms
from gpaw.kpt_descriptor import KPointDescriptor
from gpaw.lfc import BasisFunctions
from gpaw.mixer import MixerWrapper, get_mixer_from_keywords
from gpaw.mpi import MPIComm, Parallelization, world
from gpaw.new import cached_property
from gpaw.new.brillouin import BZPoints, MonkhorstPackKPoints
from gpaw.new.davidson import Davidson
from gpaw.new.density import Density
from gpaw.new.input_parameters import InputParameters
from gpaw.new.modes import FDMode, PWMode
from gpaw.new.scf import SCFLoop
from gpaw.new.smearing import OccupationNumberCalculator
from gpaw.new.symmetry import Symmetries
from gpaw.new.wave_functions import IBZWaveFunctions
from gpaw.new.xc import XCFunctional
from gpaw.setup import Setups
from gpaw.symmetry import Symmetry as OldSymmetry
from gpaw.xc import XC


class DFTComponentsBuilder:
    def __init__(self,
                 atoms: Atoms,
                 params: dict[str, Any] | InputParameters):

        self.atoms = atoms.copy()

        number_of_lattice_vectors = atoms.cell.rank
        if number_of_lattice_vectors < 3:
            raise ValueError(
                'GPAW requires 3 lattice vectors.  '
                f'Your system has {number_of_lattice_vectors}.')

        if isinstance(params, dict):
            params = InputParameters(params)
        self.params = params

        parallel = params.parallel
        world = parallel['world']

        self.mode = create_mode(**params.mode)
        self.xc = XCFunctional(XC(params.xc))  # mode?
        self.setups = Setups(atoms.numbers,
                             params.setups,
                             params.basis,
                             self.xc.setup_name,
                             world)
        self.initial_magmoms = normalize_initial_magnetic_moments(
            params.magmoms, atoms, params.spinpol)

        symmetry = create_symmetry_object(atoms,
                                          self.setups.id_a,
                                          self.initial_magmoms,
                                          params.symmetry)
        bz = create_kpts(params.kpts, atoms)
        self.ibz = symmetry.reduce(bz)

        d = parallel.get('domain', None)
        k = parallel.get('kpt', None)
        b = parallel.get('band', None)
        self.communicators = create_communicators(world, len(self.ibz),
                                                  d, k, b)
        self.grid = self.mode.create_uniform_grid(
            params.h,
            params.gpts,
            atoms.cell,
            atoms.pbc,
            symmetry,
            comm=self.communicators['d'])

        self.wf_desc = self.mode.create_wf_description(self.grid, self.dtype)
        self.nct_aX = self.mode.create_pseudo_core_densities(
            self.setups, self.grid, self.fracpos_ac)

        if self.mode.name == 'fd':
            pass  # filter = create_fourier_filter(grid)
            # setups = stups.filter(filter)

        self.nelectrons = self.setups.nvalence - params.charge

        self.nbands = calculate_number_of_bands(params.nbands,
                                                self.setups,
                                                params.charge,
                                                self.initial_magmoms,
                                                self.mode.name == 'lcao')

        self.fine_grid = self.grid.new(size=self.grid.size_c * 2)
        # decomposition=[2 * d for d in grid.decomposition]

    @cached_property
    def ncomponents(self):
        if self.initial_magmoms is None:
            return 1
        if self.initial_magmoms.ndim == 1:
            return 2
        return 4

    @cached_property
    def dtype(self):
        if self.mode.force_complex_dtype:
            return complex
        if self.ibz.bz.gamma_only:
            return float
        return complex

    @cached_property
    def fracpos_ac(self):
        return self.atoms.get_scaled_positions()

    @cached_property
    def atomdist(self):
        return self.nct_aX.layout.atomdist

    @cached_property
    def nct_R(self):
        out = self.grid.empty()
        self.nct_aX.to_uniform_grid(out=out,
                                    scale=1.0 / (self.ncomponents % 3))
        return out

    def create_potential_calculator(self):
        return self.mode.create_potential_calculator(
            self.grid,
            self.fine_grid,
            self.setups,
            self.xc,
            self.params.poissonsolver,
            self.nct_aX, self.nct_R)

    def __repr__(self):
        return f'DFTComponentsBuilder({self.atoms}, {self.params})'

    def lcao_ibz_wave_functions(self, basis_set, potential):
        from gpaw.new.lcao import create_lcao_ibz_wave_functions
        sl_default = self.params.parallel['sl_default']
        sl_lcao = self.params.parallel['sl_lcao'] or sl_default
        return create_lcao_ibz_wave_functions(
            self.setups,
            self.communicators,
            self.nbands,
            self.ncomponents,
            self.nelectrons,
            self.fracpos_ac,
            self.dtype,
            self.grid,
            self.wf_desc,
            self.ibz,
            sl_lcao,
            basis_set,
            potential)

    def random_ibz_wave_functions(self):
        return IBZWaveFunctions.from_random_numbers(
            self.ibz,
            self.communicators['b'],
            self.communicators['k'],
            self.wf_desc,
            self.setups,
            self.fracpos_ac,
            self.nbands,
            self.nelectrons,
            self.dtype)

    def create_basis_set(self):
        kd = KPointDescriptor(self.ibz.bz.kpt_Kc, self.ncomponents % 3)
        kd.set_symmetry(SimpleNamespace(pbc=self.grid.pbc),
                        self.ibz.symmetries.symmetry,
                        comm=self.communicators['w'])
        kd.set_communicator(self.communicators['k'])

        basis_set = BasisFunctions(self.grid._gd,
                                   [setup.phit_j for setup in self.setups],
                                   kd,
                                   dtype=self.dtype,
                                   cut=True)
        basis_set.set_positions(self.fracpos_ac)
        return basis_set

    def density_from_superposition(self, basis_set):
        return Density.from_superposition(self.grid,
                                          self.nct_R,
                                          self.atomdist,
                                          self.setups,
                                          basis_set,
                                          self.initial_magmoms,
                                          self.params.charge,
                                          self.params.hund)

    def create_scf_loop(self, pot_calc):
        hamiltonian = self.mode.create_hamiltonian_operator(self.wf_desc)
        eigensolver = Davidson(self.nbands,
                               self.wf_desc,
                               self.communicators['b'],
                               hamiltonian.create_preconditioner,
                               **self.params.eigensolver)

        mixer = MixerWrapper(
            get_mixer_from_keywords(self.atoms.pbc.any(),
                                    self.ncomponents, **self.params.mixer),
            self.ncomponents, self.grid._gd)

        occ_calc = OccupationNumberCalculator(
            self.params.occupations,
            self.atoms.pbc,
            self.ibz,
            self.nbands,
            self.communicators,
            self.initial_magmoms,
            self.grid.icell)

        return SCFLoop(hamiltonian, pot_calc, occ_calc,
                       eigensolver, mixer, self.communicators['w'],
                       self.params.convergence,
                       self.params.maxiter)


def create_communicators(comm: MPIComm = None,
                         nibzkpts: int = 1,
                         domain: Union[int, tuple[int, int, int]] = None,
                         kpt: int = None,
                         band: int = None) -> dict[str, MPIComm]:
    parallelization = Parallelization(comm or world, nibzkpts)
    if domain is not None:
        domain = np.prod(domain)
    parallelization.set(kpt=kpt,
                        domain=domain,
                        band=band)
    comms = parallelization.build_communicators()
    comms['w'] = comm
    return comms


def create_fourier_filter(grid):
    gamma = 1.6

    h = ((grid.icell**2).sum(1)**-0.5 / grid.size).max()

    def filter(rgd, rcut, f_r, l=0):
        gcut = np.pi / h - 2 / rcut / gamma
        ftmp = rgd.filter(f_r, rcut * gamma, gcut, l)
        f_r[:] = ftmp[:len(f_r)]

    return filter


def normalize_initial_magnetic_moments(magmoms,
                                       atoms,
                                       force_spinpol_calculation=False):
    if magmoms is None:
        magmoms = atoms.get_initial_magnetic_moments()
    elif isinstance(magmoms, float):
        magmoms = np.zeros(len(atoms)) + magmoms
    else:
        magmoms = np.array(magmoms)

    collinear = magmoms.ndim == 1
    if collinear and not magmoms.any():
        magmoms = None

    if force_spinpol_calculation and magmoms is None:
        magmoms = np.zeros(len(atoms))

    return magmoms


def create_symmetry_object(atoms, ids=None, magmoms=None, parameters=None):
    ids = ids or [()] * len(atoms)
    if magmoms is None:
        pass
    elif magmoms.ndim == 1:
        ids = [id + (m,) for id, m in zip(ids, magmoms)]
    else:
        ids = [id + tuple(m) for id, m in zip(ids, magmoms)]
    symmetry = OldSymmetry(ids, atoms.cell, atoms.pbc, **(parameters or {}))
    symmetry.analyze(atoms.get_scaled_positions())
    return Symmetries(symmetry)


def create_mode(name, **kwargs):
    if name == 'pw':
        return PWMode(**kwargs)
    if name == 'fd':
        return FDMode(**kwargs)
    1 / 0


def create_kpts(kpts: dict[str, Any], atoms: Atoms) -> BZPoints:
    if 'points' in kpts:
        assert len(kpts) == 1, kpts
        return BZPoints(kpts['points'])
    assert len(kpts) == 1
    return MonkhorstPackKPoints(kpts['size'])


def calculate_number_of_bands(nbands, setups, charge, magmoms, is_lcao):
    nao = setups.nao
    nvalence = setups.nvalence - charge
    M = 0 if magmoms is None else np.linalg.norm(magmoms.sum(0))

    orbital_free = any(setup.orbital_free for setup in setups)
    if orbital_free:
        nbands = 1

    if isinstance(nbands, str):
        if nbands == 'nao':
            nbands = nao
        elif nbands[-1] == '%':
            cfgbands = (nvalence + M) / 2
            nbands = int(np.ceil(float(nbands[:-1]) / 100 * cfgbands))
        else:
            raise ValueError('Integer expected: Only use a string '
                             'if giving a percentage of occupied bands')

    if nbands is None:
        # Number of bound partial waves:
        nbandsmax = sum(setup.get_default_nbands()
                        for setup in setups)
        nbands = int(np.ceil((1.2 * (nvalence + M) / 2))) + 4
        if nbands > nbandsmax:
            nbands = nbandsmax
        if is_lcao and nbands > nao:
            nbands = nao
    elif nbands <= 0:
        nbands = max(1, int(nvalence + M + 0.5) // 2 + (-nbands))

    if nbands > nao and is_lcao:
        raise ValueError('Too many bands for LCAO calculation: '
                         '%d bands and only %d atomic orbitals!' %
                         (nbands, nao))

    if nvalence < 0:
        raise ValueError(
            'Charge %f is not possible - not enough valence electrons' %
            charge)

    if nvalence > 2 * nbands and not orbital_free:
        raise ValueError('Too few bands!  Electrons: %f, bands: %d'
                         % (nvalence, nbands))

    return nbands



"""
==  ==========
R
r
G
g
h
x   r or h
==  ==========

"""

from __future__ import annotations

from collections import defaultdict

import numpy as np

from gpaw.core import PlaneWaves, UniformGrid
from gpaw.core.uniform_grid import UniformGridFunctions
from gpaw.new.potential import Potential
from gpaw.new.xc import XCFunctional
from gpaw.setup import Setup
from gpaw.typing import Array1D, Array3D
from gpaw.utilities import pack, unpack


class PotentialCalculator:
    def __init__(self,
                 xc: XCFunctional,
                 poisson_solver,
                 setups: list[Setup],
                 nct_R: UniformGridFunctions):
        self.poisson_solver = poisson_solver
        self.xc = xc
        self.setups = setups
        self.nct_R = nct_R

    def __str__(self):
        return f'\n{self.poisson_solver}\n{self.xc}'

    def calculate(self, density, vHt_x=None):
        energies, vt_sR, vHt_x = self._calculate(density, vHt_x)

        Q_aL = self.calculate_charges(vHt_x)
        dH_asii, corrections = calculate_non_local_potential(
            self.setups, density, self.xc, Q_aL)

        for key, e in corrections.items():
            # print(key, e, energies[key])
            energies[key] += e

        return Potential(vt_sR, dH_asii, energies), vHt_x, Q_aL

    def move(self, fracpos_ac, ndensities):
        delta_nct_R = self.nct_R.new()
        delta_nct_R.data = -self.nct_R.data
        self._move(fracpos_ac, ndensities)
        delta_nct_R.data += self.nct_R.data
        return delta_nct_R


class UniformGridPotentialCalculator(PotentialCalculator):
    def __init__(self,
                 wf_grid: UniformGrid,
                 fine_grid: UniformGrid,
                 setups,
                 xc,
                 poisson_solver,
                 nct_aR, nct_R):
        self.nct_aR = nct_aR

        fracpos_ac = nct_aR.fracpos_ac

        self.vbar_ar = setups.create_local_potentials(fine_grid, fracpos_ac)
        self.ghat_aLr = setups.create_compensation_charges(fine_grid,
                                                           fracpos_ac)

        self.vbar_r = fine_grid.empty()
        self.vbar_ar.to_uniform_grid(out=self.vbar_r)

        self.interpolate = wf_grid.transformer(fine_grid)
        self.restrict = fine_grid.transformer(wf_grid)

        super().__init__(xc, poisson_solver, setups, nct_R)

    def calculate_charges(self, vHt_r):
        return self.ghat_aLr.integrate(vHt_r)

    def _calculate(self, density, vHt_r):
        nt_sR = density.nt_sR
        nt_sr = self.interpolate(nt_sR)
        if not nt_sR.desc.pbc_c.all():
            scale_s = nt_sR.integrate() / nt_sr.integrate()
            for scale, nt_r in zip(scale_s, nt_sr):
                nt_r.data *= scale
        grid2 = nt_sr.desc

        vxct_sr = grid2.zeros(nt_sr.dims)
        e_xc = self.xc.calculate(nt_sr, vxct_sr)

        charge_r = grid2.empty()
        charge_r.data[:] = nt_sr.data[:density.ndensities].sum(axis=0)
        e_zero = self.vbar_r.integrate(charge_r)

        ccc_aL = density.calculate_compensation_charge_coefficients()
        self.ghat_aLr.add_to(charge_r, ccc_aL)
        if vHt_r is None:
            vHt_r = grid2.zeros()
        self.poisson_solver.solve(vHt_r, charge_r)
        e_coulomb = 0.5 * vHt_r.integrate(charge_r)

        vt_sr = vxct_sr
        vt_sr.data += vHt_r.data + self.vbar_r.data
        vt_sR = self.restrict(vt_sr)
        e_kinetic = 0.0
        for spin, (vt_R, nt_R) in enumerate(zip(vt_sR, nt_sR)):
            e_kinetic -= vt_R.integrate(nt_R)
            if spin < density.ndensities:
                e_kinetic += vt_R.integrate(self.nct_R)

        e_external = 0.0

        return {'kinetic': e_kinetic,
                'coulomb': e_coulomb,
                'zero': e_zero,
                'xc': e_xc,
                'external': e_external}, vt_sR, vHt_r

    def _move(self, fracpos_ac, ndensities):
        self.ghat_aLr.move(fracpos_ac)
        self.vbar_ar.move(fracpos_ac)
        self.vbar_ar.to_uniform_grid(out=self.vbar_r)
        self.nct_aR.move(fracpos_ac)
        self.nct_aR.to_uniform_grid(out=self.nct_R, scale=1.0 / ndensities)

    def force_contributions(self, state):
        density = state.density
        potential = state.potential
        nt_R = density.nt_sR[0]
        vt_R = potential.vt_sR[0]
        if density.ndensities > 1:
            nt_R = nt_R.desc.empty()
            nt_R.data[:] = density.nt_sR.data[:density.ndensities].sum(axis=0)
            vt_R = vt_R.desc.empty()
            vt_R.data[:] = (
                potential.vt_sR.data[:density.ndensities].sum(axis=0) /
                density.ndensities)

        nt_r = self.interpolate(nt_R)
        if not nt_r.desc.pbc_c.all():
            scale = nt_R.integrate() / nt_r.integrate()
            nt_r.data *= scale

        return (self.ghat_aLr.derivative(state.vHt_x),
                self.nct_aR.derivative(vt_R),
                self.vbar_ar.derivative(nt_r))


class PlaneWavePotentialCalculator(PotentialCalculator):
    def __init__(self,
                 grid,
                 fine_grid,
                 pw: PlaneWaves,
                 fine_pw: PlaneWaves,
                 setups,
                 xc,
                 poisson_solver,
                 nct_ag,
                 nct_R):
        super().__init__(xc, poisson_solver, setups, nct_R)

        fracpos_ac = nct_ag.fracpos_ac
        self.nct_ag = nct_ag
        self.vbar_ag = setups.create_local_potentials(pw, fracpos_ac)
        self.ghat_aLh = setups.create_compensation_charges(fine_pw, fracpos_ac)

        self.h_g = fine_pw.map_indices(pw)
        self.fftplan, self.ifftplan = grid.fft_plans()
        self.fftplan2, self.ifftplan2 = fine_grid.fft_plans()
        self.fine_grid = fine_grid

        self.vbar_g = pw.zeros()
        self.vbar_ag.add_to(self.vbar_g)

    def calculate_charges(self, vHt_h):
        return self.ghat_aLh.integrate(vHt_h)

    def _calculate(self, density, vHt_h):
        nt_sr = self.fine_grid.empty(density.nt_sR.dims)
        nt_g = self.vbar_g.desc.zeros()
        indices = nt_g.desc.indices(self.fftplan.out_R.shape)
        for spin, (nt_R, nt_r) in enumerate(zip(density.nt_sR, nt_sr)):
            nt_R.fft_interpolate(nt_r, self.fftplan, self.ifftplan2)
            if spin < density.ndensities:
                nt_g.data += self.fftplan.out_R.ravel()[indices]
        nt_g.data *= 1 / self.fftplan.in_R.size

        e_zero = self.vbar_g.integrate(nt_g)

        if vHt_h is None:
            vHt_h = self.ghat_aLh.pw.zeros()

        charge_h = vHt_h.desc.zeros()
        coef_aL = density.calculate_compensation_charge_coefficients()
        self.ghat_aLh.add_to(charge_h, coef_aL)
        charge_h.data[self.h_g] += nt_g.data
        # background charge ???

        self.poisson_solver.solve(vHt_h, charge_h)
        e_coulomb = 0.5 * vHt_h.integrate(charge_h)

        vt_g = self.vbar_g.copy()
        vt_g.data += vHt_h.data[self.h_g]

        vt_sR = density.nt_sR.new()
        vt_sR.data[:] = vt_g.ifft(self.ifftplan, grid=vt_sR.desc).data
        vxct_sr = nt_sr.desc.zeros(density.nt_sR.dims)
        e_xc = self.xc.calculate(nt_sr, vxct_sr)

        vtmp_R = vt_sR.desc.empty()
        e_kinetic = 0.0
        for spin, (vt_R, vxct_r) in enumerate(zip(vt_sR, vxct_sr)):
            vxct_r.fft_restrict(vtmp_R, self.fftplan2, self.ifftplan)
            vt_R.data += vtmp_R.data
            e_kinetic -= vt_R.integrate(density.nt_sR[spin])
            if spin < density.ndensities:
                e_kinetic += vt_R.integrate(self.nct_R)

        e_external = 0.0

        return {'kinetic': e_kinetic,
                'coulomb': e_coulomb,
                'zero': e_zero,
                'xc': e_xc,
                'external': e_external}, vt_sR, vHt_h

    def _move_nct(self, fracpos_ac, ndensities):
        self.ghat_aLr.move(fracpos_ac)
        self.vbar_ar.move(fracpos_ac)
        self.vbar_ar.to_uniform_grid(out=self.vbar_r)
        self.nct_aR.move(fracpos_ac)
        self.nct_aR.to_uniform_grid(out=self.nct_R, scale=1.0 / ndensities)

    def forces(self, nct_ag):
        return (self.ghat_ah.derivative(self.vHt_h),
                nct_ag.derivative(self.vt_g),
                self.vbar_ag.derivative(self.nt_g))


def calculate_non_local_potential(setups,
                                  density,
                                  xc,
                                  Q_aL):
    dH_asii = density.D_asii.new()
    energy_corrections = defaultdict(float)
    for a, D_sii in density.D_asii.items():
        Q_L = Q_aL[a]
        setup = setups[a]
        dH_sii, energies = calculate_non_local_potential1(
            setup, xc, D_sii, Q_L)
        dH_asii[a][:] = dH_sii
        for key, e in energies.items():
            energy_corrections[key] += e

    # Sum over domain:
    names = ['kinetic',
             'coulomb',
             'zero',
             'xc',
             'external']
    energies = np.array([energy_corrections[name] for name in names])
    density.D_asii.layout.atomdist.comm.sum(energies)
    energy_corrections = {name: e for name, e in zip(names, energies)}
    return dH_asii, energy_corrections


def calculate_non_local_potential1(setup: Setup,
                                   xc: XCFunctional,
                                   D_sii: Array3D,
                                   Q_L: Array1D) -> tuple[Array3D,
                                                          dict[str, float]]:
    ndensities = 2 if len(D_sii) == 2 else 1
    D_sp = np.array([pack(D_ii) for D_ii in D_sii])

    D_p = D_sp[:ndensities].sum(0)

    dH_p = (setup.K_p + setup.M_p +
            setup.MB_p + 2.0 * setup.M_pp @ D_p +
            setup.Delta_pL @ Q_L)
    e_kinetic = setup.K_p @ D_p + setup.Kc
    e_zero = setup.MB + setup.MB_p @ D_p
    e_coulomb = setup.M + D_p @ (setup.M_p + setup.M_pp @ D_p)

    dH_sp = np.zeros_like(D_sp)
    dH_sp[:ndensities] = dH_p
    e_xc = xc.calculate_paw_correction(setup, D_sp, dH_sp)
    e_kinetic -= (D_sp * dH_sp).sum().real

    e_external = 0.0

    dH_sii = unpack(dH_sp)

    return dH_sii, {'kinetic': e_kinetic,
                    'coulomb': e_coulomb,
                    'zero': e_zero,
                    'xc': e_xc,
                    'external': e_external}



from __future__ import annotations

import numpy as np

from gpaw.core.arrays import DistributedArrays
from gpaw.mpi import MPIComm, serial_comm
from gpaw.typing import Array1D


class AtomArraysLayout:
    def __init__(self,
                 shapes: list[int | tuple[int, ...]],
                 atomdist: AtomDistribution | MPIComm = serial_comm,
                 dtype=float):
        self.shape_a = [shape if isinstance(shape, tuple) else (shape,)
                        for shape in shapes]
        if not isinstance(atomdist, AtomDistribution):
            atomdist = AtomDistribution(np.zeros(len(shapes), int), atomdist)
        self.atomdist = atomdist
        self.dtype = np.dtype(dtype)

        self.size = sum(np.prod(shape) for shape in self.shape_a)

        self.myindices = []
        self.mysize = 0
        I1 = 0
        for a in atomdist.indices:
            I2 = I1 + np.prod(self.shape_a[a])
            self.myindices.append((a, I1, I2))
            self.mysize += I2 - I1
            I1 = I2

    def __repr__(self):
        return (f'AtomArraysLayout({self.shape_a}, {self.atomdist}, '
                f'{self.dtype})')

    def new(self, atomdist=None):
        return AtomArraysLayout(self.shape_a, atomdist or self.atomdist,
                                self.dtype)

    def empty(self,
              dims: int | tuple[int, ...] = (),
              comm: MPIComm = serial_comm,
              transposed=False) -> AtomArrays:
        return AtomArrays(self, dims, comm, transposed=transposed)


class AtomDistribution:
    def __init__(self, ranks, comm):
        self.comm = comm
        self.rank_a = ranks
        self.indices = np.where(ranks == comm.rank)[0]

    def __repr__(self):
        return (f'AtomDistribution(ranks={self.rank_a}, '
                f'comm={self.comm.rank}/{self.comm.size})')


class AtomArrays(DistributedArrays):
    def __init__(self,
                 layout: AtomArraysLayout,
                 dims: int | tuple[int, ...] = (),
                 comm: MPIComm = serial_comm,
                 data: np.ndarray = None,
                 transposed=False):
        DistributedArrays. __init__(self, dims, (layout.mysize,),
                                    comm, layout.atomdist.comm,
                                    dtype=layout.dtype,
                                    data=data,
                                    dv=np.nan,
                                    transposed=transposed)
        self.layout = layout
        self._arrays = {}
        for a, I1, I2 in layout.myindices:
            if transposed:
                self._arrays[a] = self.data[I1:I2].reshape(
                    layout.shape_a[a] + self.mydims)
            else:
                self._arrays[a] = self.data[..., I1:I2].reshape(
                    self.mydims + layout.shape_a[a])
        self.natoms: int = len(layout.shape_a)

    def __repr__(self):
        return f'AtomArrays({self.layout})'

    def new(self, layout=None, data=None):
        return AtomArrays(layout or self.layout,
                          self.dims,
                          self.comm,
                          data=data,
                          transposed=self.transposed)

    def __getitem__(self, a):
        return self._arrays[a]

    def get(self, a):
        return self._arrays.get(a)

    def __setitem__(self, a, value):
        self._arrays[a][:] = value

    def __contains__(self, a):
        return a in self._arrays

    def items(self):
        return self._arrays.items()

    def keys(self):
        return self._arrays.keys()

    def values(self):
        return self._arrays.values()

    def gather(self, broadcast=False, copy=False) -> AtomArrays | None:
        assert not self.transposed
        comm = self.layout.atomdist.comm
        if comm.size == 1:
            if copy:
                aa = self.new()
                aa.data[:] = self.data
                return aa
            return self

        if comm.rank == 0 or broadcast:
            aa = self.new(layout=self.layout.new(atomdist=serial_comm))
        else:
            aa = None

        if comm.rank == 0:
            size_ra, size_r = self.sizes()
            shape = self.mydims + (size_r.max(),)
            buffer = np.empty(shape, self.layout.dtype)
            for rank in range(1, comm.size):
                buf = buffer[..., :size_r[rank]]
                comm.receive(buf, rank)
                b1 = 0
                for a, size in size_ra[rank].items():
                    b2 = b1 + size
                    aa[a] = buf[..., b1:b2].reshape(self.myshape +
                                                    self.layout.shape_a[a])
            for a, array in self._arrays.items():
                aa[a] = array
        else:
            comm.send(self.data, 0)

        if broadcast:
            comm.broadcast(aa.data, 0)

        return aa

    def sizes(self) -> tuple[list[dict[int, int]], Array1D]:
        comm = self.layout.atomdist.comm
        size_ra: list[dict[int, int]] = [{} for _ in range(comm.size)]
        size_r = np.zeros(comm.size, int)
        for a, (rank, shape) in enumerate(zip(self.layout.atomdist.rank_a,
                                              self.layout.shape_a)):
            size = np.prod(shape)
            size_ra[rank][a] = size
            size_r[rank] += size
        return size_ra, size_r

    def _dict_view(self):
        if self.transposed:
            return {a: np.moveaxis(array, 0, -1)
                    for a, array in self._arrays.items()}
        return self

    def scatter_from(self, data: np.ndarray = None) -> None:
        assert not self.transposed
        comm = self.layout.atomdist.comm
        if comm.size == 1:
            self.data[:] = data
            return

        if comm.rank != 0:
            comm.receive(self.data, 0, 42)
            return

        size_ra, size_r = self.sizes()
        aa = self.new(layout=self.layout.new(atomdist=serial_comm),
                      data=data)
        requests = []
        for rank, (totsize, size_a) in enumerate(zip(size_r, size_ra)):
            if rank != 0:
                buf = np.empty(self.mydims + (totsize,), self.layout.dtype)
                b1 = 0
                for a, size in size_a.items():
                    b2 = b1 + size
                    buf[..., b1:b2] = aa[a].reshape(self.mydims + (size,))
                request = comm.send(buf, rank, 42, False)
                # Remember to store a reference to the
                # send buffer (buf) so that is isn't
                # deallocated
                requests.append((request, buf))
            else:
                for a in size_a:
                    self[a] = aa[a]

        for request, _ in requests:
            comm.wait(request)



import numpy as np
import pytest

from gpaw.mpi import world
from gpaw.atom.radialgd import EquidistantRadialGridDescriptor
from gpaw.fd_operators import Gradient
from gpaw.grid_descriptor import GridDescriptor
from gpaw.setup import Setup, Setups
from gpaw.spherical_harmonics import YL
from gpaw.utilities.tools import coordinates


def rlYlm(L, r_vg):
    r"""Calculates :math:`r^{l} Y_{l,m}(\theta, \varphi)` on grid."""
    rlYlm_g = np.zeros_like(r_vg[0])
    for c, n_v in YL[L]:
        rlYlm_g += c * np.prod(np.moveaxis(r_vg, 0, -1)**n_v, -1)
    return rlYlm_g


class DummySetup(Setup):
    def __init__(self, lmax):
        self.l_j = range(lmax + 1)
        self.nj = lmax + 1
        self.ni = (lmax + 1)**2


def calculate_integrals_on_regular_grid(radial_function, *,
                                        lmax=2,
                                        cell_v=[24, 24, 24],
                                        N_v=[128, 128, 128]):
    cell_v = np.asarray(cell_v, dtype=float)
    gd = GridDescriptor(N_v, cell_v, False)
    origin_v = 0.5 * cell_v
    r_vg, r2_g = coordinates(gd, origin=origin_v)
    r_g = np.sqrt(r2_g)
    radial_g = radial_function(r_g)

    grad_v = []
    for v in range(3):
        grad_v.append(Gradient(gd, v, n=3))
    grad_phi2_vg = gd.empty(3)

    Lmax = (lmax + 1)**2  # 1+3+5+...
    nabla_LLv = np.zeros((Lmax, Lmax, 3))
    rxnabla_LLv = np.zeros((Lmax, Lmax, 3))
    for L2 in range(Lmax):
        phi2_g = radial_g * rlYlm(L2, r_vg)
        for v in range(3):
            grad_v[v].apply(phi2_g, grad_phi2_vg[v])
        for L1 in range(Lmax):
            phi1_g = radial_g * rlYlm(L1, r_vg)

            def nabla(v):
                return gd.integrate(phi1_g * grad_phi2_vg[v])

            def rxnabla(v1, v2):
                return gd.integrate(phi1_g *
                                    (r_vg[v1] * grad_phi2_vg[v2] -
                                     r_vg[v2] * grad_phi2_vg[v1]))

            for v in range(3):
                v1 = (v + 1) % 3
                v2 = (v + 2) % 3
                nabla_LLv[L1, L2, v] = nabla(v)
                rxnabla_LLv[L1, L2, v] = rxnabla(v1, v2)

    return {'nabla': nabla_LLv, 'rxnabla': rxnabla_LLv}


def calculate_integrals_on_radial_grid(radial_function, *,
                                       lmax=2,
                                       h=1e-3, N=12e3,
                                       use_phit=False,
                                       alt_rxnabla=False):
    setup = DummySetup(lmax)
    rgd = EquidistantRadialGridDescriptor(h, int(N))
    r_g = rgd.r_g
    radial_g = radial_function(r_g)
    phi_jg = [radial_g * r_g**l for l in setup.l_j]
    phit_jg = np.zeros_like(phi_jg)
    if use_phit:
        phit_jg, phi_jg = phi_jg, phit_jg
    nabla_LLv = setup.get_derivative_integrals(rgd, phi_jg, phit_jg)
    if alt_rxnabla:
        from .rxnabla_alternative import get_magnetic_integrals_alt
        rxnabla_LLv = get_magnetic_integrals_alt(setup, rgd, phi_jg, phit_jg)
    else:
        rxnabla_LLv = setup.get_magnetic_integrals(rgd, phi_jg, phit_jg)
    return {'nabla': nabla_LLv, 'rxnabla': rxnabla_LLv}


@pytest.fixture(scope='module')
def lmax():
    return 3


@pytest.fixture(scope='module')
def radial_function():
    return lambda r_g: np.exp(-0.1 * r_g**2)


@pytest.fixture(scope='module')
def integrals_on_regular_grid(lmax, radial_function):
    # Increase accuracy with the number of processes
    # Here, world.size = 40 triggers more accurate calculation
    N = {2: 48, 4: 64, 8: 92, 40: 256}.get(world.size, 32)
    return calculate_integrals_on_regular_grid(radial_function,
                                               lmax=lmax,
                                               N_v=[N, N, N])


@pytest.fixture(scope='module')
def integrals_on_radial_grid(lmax, radial_function):
    return calculate_integrals_on_radial_grid(radial_function,
                                              lmax=lmax)


@pytest.mark.parametrize('kind', ['nabla', 'rxnabla'])
@pytest.mark.parametrize('lmax_test', range(4))
def test_integrals(kind,
                   lmax,
                   lmax_test,
                   integrals_on_regular_grid,
                   integrals_on_radial_grid):
    # This test is parametrized over different lmax values,
    # because the accuracy of the regular grid calculation
    # depends strongly on that
    assert lmax >= lmax_test, 'this test is broken, increase lmax'
    arr1_LLv = integrals_on_regular_grid[kind]
    arr2_LLv = integrals_on_radial_grid[kind]
    if world.rank == 0 and lmax_test == lmax:
        np.set_printoptions(precision=4, suppress=True, linewidth=2000)
        print(kind)
        for v in range(3):
            print('xyz'[v])
            print(arr1_LLv[..., v])
            print(arr2_LLv[..., v])
    world.barrier()
    # Accuracy is increased with the finer grid
    # (as triggered by the number of processes)
    rtol = {2: 5e-5, 4: 9e-6, 8: 1e-6, 40: 6e-8}.get(world.size, 1e-3)
    atol = 1e-12
    if lmax_test == 3:
        # Regular grid is not fine enough for high lmax, which
        # results in large spurious values that should be zero
        atol = {2: 3e-1, 4: 4e-2, 8: 5e-3, 40: 2e-5}.get(world.size, 3e-0)
    Lmax = (lmax_test + 1)**2
    assert np.allclose(arr1_LLv[:Lmax, :Lmax], arr2_LLv[:Lmax, :Lmax],
                       rtol=rtol, atol=atol)


def test_phit_integrals(lmax, radial_function, integrals_on_radial_grid):
    phit_integrals = \
        calculate_integrals_on_radial_grid(radial_function,
                                           lmax=lmax,
                                           use_phit=True)
    for kind, ref_LLv in integrals_on_radial_grid.items():
        arr_LLv = phit_integrals[kind]
        assert np.allclose(ref_LLv, -arr_LLv, rtol=0, atol=0)


@pytest.mark.parametrize('kind', ['nabla', 'rxnabla'])
def test_skew_symmetry(kind, integrals_on_radial_grid):
    arr_LLv = integrals_on_radial_grid[kind]
    rtol = {'nabla': 1e-8, 'rxnabla': 0}[kind]
    for v in range(3):
        arr_LL = arr_LLv[..., v]
        assert np.allclose(arr_LL, -arr_LL.T, rtol=rtol, atol=1e-11)


def test_rxnabla_vs_alt_implementation(lmax, radial_function,
                                       integrals_on_radial_grid):
    # Alternative implementation supports lmax upto 2
    lmax = min(lmax, 2)
    Lmax = (lmax + 1)**2
    rxnabla_LLv = integrals_on_radial_grid['rxnabla'][:Lmax, :Lmax, :]
    alt_LLv = calculate_integrals_on_radial_grid(radial_function,
                                                 lmax=lmax,
                                                 alt_rxnabla=True)['rxnabla']
    assert np.allclose(rxnabla_LLv, alt_LLv, rtol=0, atol=1e-12)


def test_lmax_zero(radial_function):
    integrals = \
        calculate_integrals_on_radial_grid(radial_function,
                                           lmax=0)
    for kind, ref_LLv in integrals.items():
        assert np.allclose(ref_LLv, 0, rtol=0, atol=0)


@pytest.mark.parametrize('xc', ['LDA', 'PBE'])
@pytest.mark.parametrize('Z', (
    list(range(1, 43)) + list(range(44, 57)) + list(range(72, 84)) + [86]))
def test_skew_symmetry_real_setup(Z, xc):
    setup = Setups([Z], {}, {}, xc)[0]
    for v in range(3):
        arr_ii = setup.nabla_iiv[:, :, v]
        assert np.allclose(arr_ii, -arr_ii.T, rtol=5e-6, atol=1e-12)
        arr_ii = setup.rxnabla_iiv[:, :, v]
        assert np.allclose(arr_ii, -arr_ii.T, rtol=0, atol=1e-14)



from __future__ import annotations
from typing import TYPE_CHECKING
import numpy as np
from gpaw.core.atom_arrays import (AtomArrays, AtomArraysLayout,
                                   AtomDistribution)
from gpaw.mpi import serial_comm, MPIComm
from gpaw.kpt_descriptor import KPointDescriptor
from gpaw.lfc import LocalizedFunctionsCollection as LFC
from gpaw.spline import Spline
from gpaw.typing import ArrayLike2D
if TYPE_CHECKING:
    from gpaw.core.uniform_grid import UniformGridFunctions


def to_spline(l, rcut, f):
    r = np.linspace(0, rcut, 100)
    return Spline(l, rcut, f(r))


class AtomCenteredFunctions:
    def __init__(self,
                 functions,
                 fracpos_ac: ArrayLike2D):
        self.functions = [[to_spline(*f) if isinstance(f, tuple) else f
                           for f in funcs]
                          for funcs in functions]
        self.fracpos_ac = np.array(fracpos_ac)

        self._layout = None
        self._lfc = None

    def __repr__(self):
        funcs = [['spdfgh'[f.l] for f in ff] for ff in self.functions[:4]]
        if len(self.functions) > 4:
            funcs.append(...)
        return f'{self.__class__.__name__}(functions={funcs})'

    @property
    def layout(self):
        self._lacy_init()
        return self._layout

    def empty(self,
              dims: int | tuple[int, ...] = (),
              comm: MPIComm = serial_comm,
              transposed=False) -> AtomArrays:
        return self.layout.empty(dims, comm, transposed=transposed)

    def move(self, fracpos_ac):
        self.fracpos_ac = np.array(fracpos_ac)
        self._lfc.set_positions(fracpos_ac)

    def add_to(self, functions, coefs=1.0):
        self._lacy_init()

        if isinstance(coefs, float):
            self._lfc.add(functions.data, coefs)
            return

        self._lfc.add(functions.data, coefs._dict_view(), q=0)

    def integrate(self, functions, out=None):
        self._lacy_init()
        if out is None:
            out = self.layout.empty(functions.dims, functions.comm)
        self._lfc.integrate(functions.data, out._dict_view(), q=0)
        return out

    def derivative(self, functions, out=None):
        self._lacy_init()
        if out is None:
            out = self.layout.empty(functions.dims + (3,), functions.comm,
                                    transposed=True)
        else:
            assert out.transposed
        coef_axiv = {a: np.moveaxis(array_ixv, 0, -2)
                     for a, array_ixv in out._arrays.items()}
        self._lfc.derivative(functions.data, coef_axiv, q=0)
        return out


class UniformGridAtomCenteredFunctions(AtomCenteredFunctions):
    def __init__(self, functions, fracpos_ac, grid, integral=None, cut=False):
        AtomCenteredFunctions.__init__(self, functions, fracpos_ac)
        self.grid = grid
        self.integral = integral
        self.cut = cut

    def _lacy_init(self):
        if self._lfc is not None:
            return
        gd = self.grid._gd
        kd = KPointDescriptor(np.array([self.grid.kpt]))
        self._lfc = LFC(gd, self.functions, kd,
                        dtype=self.grid.dtype,
                        integral=self.integral,
                        forces=True,
                        cut=self.cut)
        self._lfc.set_positions(self.fracpos_ac)
        atomdist = AtomDistribution(
            ranks=np.array([sphere.rank for sphere in self._lfc.sphere_a]),
            comm=self.grid.comm)
        self._layout = AtomArraysLayout([sum(2 * f.l + 1 for f in funcs)
                                         for funcs in self.functions],
                                        atomdist,
                                        self.grid.dtype)

    def to_uniform_grid(self,
                        out: UniformGridFunctions,
                        scale: float = 1.0) -> UniformGridFunctions:
        out.data[:] = 0.0
        self.add_to(out, scale)
        return out



from __future__ import annotations
import itertools
import warnings
from functools import partial
from math import inf
from types import SimpleNamespace
from typing import TYPE_CHECKING

from gpaw.scf import dict2criterion, write_iteration
if TYPE_CHECKING:
    from gpaw.new.calculation import DFTState


class SCFConvergenceError(Exception):
    ...


class SCFLoop:
    def __init__(self,
                 hamiltonian,
                 pot_calc,
                 occ_calc,
                 eigensolver,
                 mixer,
                 world,
                 convergence,
                 maxiter):
        self.hamiltonian = hamiltonian
        self.pot_calc = pot_calc
        self.eigensolver = eigensolver
        self.mixer = mixer
        self.occ_calc = occ_calc
        self.world = world
        self.convergence = convergence
        self.maxiter = maxiter

    def __str__(self):
        return str(self.pot_calc)

    def iterate(self,
                state: DFTState,
                convergence=None,
                maxiter=None,
                log=None):
        cc = create_convergence_criteria(convergence or self.convergence)
        maxiter = maxiter or self.maxiter

        dS = state.density.overlap_correction

        self.mixer.reset()

        dens_error = inf  # ???
        dens_error = self.mixer.mix(state.density)

        for niter in itertools.count(start=1):
            dH = state.potential.dH
            Ht = partial(self.hamiltonian.apply, state.potential.vt_sR)
            wfs_error = self.eigensolver.iterate(state.ibzwfs, Ht, dH, dS)
            state.ibzwfs.calculate_occs(self.occ_calc)

            ctx = SCFContext(
                state, niter,
                wfs_error, dens_error,
                self.world)

            yield ctx

            entries, converged = check_convergence(ctx, cc)
            if log:
                write_iteration(cc, converged, entries, ctx, log)
            if all(converged.values()):
                break
            if niter == maxiter:
                raise SCFConvergenceError

            state.density.update(self.pot_calc.nct_R, state.ibzwfs)
            dens_error = self.mixer.mix(state.density)
            state.potential, state.vHt_x, _ = self.pot_calc.calculate(
                state.density, state.vHt_x)


class SCFContext:
    def __init__(self,
                 state: DFTState,
                 niter: int,
                 wfs_error: float,
                 dens_error: float,
                 world):
        self.state = state
        self.niter = niter
        energy = (sum(state.potential.energies.values()) +
                  sum(state.ibzwfs.energies.values()))
        self.ham = SimpleNamespace(e_total_extrapolated=energy)
        self.wfs = SimpleNamespace(nvalence=state.ibzwfs.nelectrons,
                                   world=world,
                                   eigensolver=SimpleNamespace(
                                       error=wfs_error),
                                   nspins=state.density.ndensities,
                                   collinear=state.density.collinear)
        self.dens = SimpleNamespace(
            calculate_magnetic_moments=state.density
            .calculate_magnetic_moments,
            fixed=False,
            error=dens_error)


def check_convergence(ctx, criteria):
    entries = {}  # for log file, per criteria
    converged_items = {}  # True/False, per criteria

    for name, criterion in criteria.items():
        if not criterion.calc_last:
            converged_items[name], entries[name] = criterion(ctx)

    converged = all(converged_items.values())

    for name, criterion in criteria.items():
        if criterion.calc_last:
            if converged:
                converged_items[name], entries[name] = criterion(ctx)
            else:
                converged_items[name], entries[name] = False, ''

    # Converged?
    return entries, converged_items


def create_convergence_criteria(criteria):
    # Gather convergence criteria for SCF loop.
    custom = criteria.pop('custom', [])
    for name, criterion in criteria.items():
        if hasattr(criterion, 'todict'):
            # 'Copy' so no two calculators share an instance.
            criteria[name] = dict2criterion(criterion.todict())
        else:
            criteria[name] = dict2criterion({name: criterion})

    if not isinstance(custom, (list, tuple)):
        custom = [custom]
    for criterion in custom:
        if isinstance(criterion, dict):  # from .gpw file
            msg = ('Custom convergence criterion "{:s}" encountered, '
                   'which GPAW does not know how to load. This '
                   'criterion is NOT enabled; you may want to manually'
                   ' set it.'.format(criterion['name']))
            warnings.warn(msg)
            continue

        criteria[criterion.name] = criterion
        msg = ('Custom convergence criterion {:s} encountered. '
               'Please be sure that each calculator is fed a '
               'unique instance of this criterion. '
               'Note that if you save the calculator instance to '
               'a .gpw file you may not be able to re-open it. '
               .format(criterion.name))
        warnings.warn(msg)

    for criterion in criteria.values():
        criterion.reset()

    return criteria



from __future__ import annotations

from typing import Any

import numpy as np
from ase.units import Bohr, Ha
from gpaw.core import PlaneWaves, UniformGrid
from gpaw.fd_operators import Laplace
from gpaw.mpi import MPIComm, serial_comm
from gpaw.new.poisson import (PoissonSolver, PoissonSolverWrapper,
                              ReciprocalSpacePoissonSolver)
from gpaw.new.pot_calc import (PlaneWavePotentialCalculator,
                               UniformGridPotentialCalculator)
from gpaw.poisson import PoissonSolver as make_poisson_solver
from gpaw.utilities.gpts import get_number_of_grid_points
import _gpaw


class Mode:
    name: str
    interpolation: str

    def __init__(self, force_complex_dtype=False):
        self.force_complex_dtype = force_complex_dtype

    def create_uniform_grid(self,
                            h: float | None,
                            gpts,
                            cell,
                            pbc,
                            symmetry,
                            comm: MPIComm = serial_comm) -> UniformGrid:
        cell = cell / Bohr
        if h is not None:
            h /= Bohr

        realspace = (self.name != 'pw' and self.interpolation != 'fft')
        if not realspace:
            pbc = (True, True, True)

        if gpts is not None:
            if h is not None:
                raise ValueError("""You can't use both "gpts" and "h"!""")
            size = gpts
        else:
            size = get_number_of_grid_points(cell, h, self, realspace,
                                             symmetry.symmetry)
        return UniformGrid(cell=cell, pbc=pbc, size=size, comm=comm)


class PWMode(Mode):
    name = 'pw'

    def __init__(self, ecut: float = 340.0):
        Mode.__init__(self)
        if ecut is not None:
            ecut /= Ha
        self.ecut = ecut

    def create_wf_description(self,
                              grid: UniformGrid,
                              dtype) -> PlaneWaves:

        return PlaneWaves(ecut=self.ecut, cell=grid.cell, dtype=dtype)

    def create_pseudo_core_densities(self, setups, grid, fracpos_ac):
        pw = PlaneWaves(ecut=2 * self.ecut, cell=grid.cell)
        return setups.create_pseudo_core_densities(pw, fracpos_ac)

    def create_poisson_solver(self, fine_grid_pw, params):
        return ReciprocalSpacePoissonSolver(fine_grid_pw)

    def create_potential_calculator(self,
                                    grid,
                                    fine_grid,
                                    setups,
                                    xc,
                                    poisson_solver_params,
                                    nct_ag, nct_R):
        pw = nct_ag.pw
        fine_pw = pw.new(ecut=8 * self.ecut)
        poisson_solver = self.create_poisson_solver(fine_pw,
                                                    poisson_solver_params)
        return PlaneWavePotentialCalculator(grid,
                                            fine_grid,
                                            pw,
                                            fine_pw,
                                            setups,
                                            xc,
                                            poisson_solver,
                                            nct_ag, nct_R)

    def create_hamiltonian_operator(self, grid, blocksize=10):
        return PWHamiltonian()


class PWHamiltonian:
    def apply(self, vt_sR, psit_nG, out, spin):
        out_nG = out
        vt_R = vt_sR.data[spin]
        np.multiply(psit_nG.desc.ekin_G, psit_nG.data, out_nG.data)
        f_R = None
        for p_G, o_G in zip(psit_nG, out_nG):
            f_R = p_G.ifft(grid=vt_sR.desc, out=f_R)
            f_R.data *= vt_R
            o_G.data += f_R.fft(pw=p_G.desc).data
        return out_nG

    def create_preconditioner(self, blocksize):
        return precondition


def precondition(psit, residuals, out):
    G2 = psit.desc.ekin_G * 2
    for r, o, ekin in zip(residuals.data, out.data, psit.norm2('kinetic')):
        _gpaw.pw_precond(G2, r, ekin, o)


class FDMode(Mode):
    name = 'fd'
    stencil = 3
    interpolation = 'not fft'

    def create_wf_description(self,
                              grid: UniformGrid,
                              dtype) -> UniformGrid:
        return grid.new(dtype=dtype)

    def create_pseudo_core_densities(self, setups, grid, fracpos_ac):
        return setups.create_pseudo_core_densities(grid, fracpos_ac)

    def create_poisson_solver(self,
                              grid: UniformGrid,
                              params: dict[str, Any]) -> PoissonSolver:
        solver = make_poisson_solver(**params)
        solver.set_grid_descriptor(grid._gd)
        return PoissonSolverWrapper(solver)

    def create_potential_calculator(self, grid, fine_grid,
                                    setups,
                                    xc,
                                    poisson_solver_params,
                                    nct_aR, nct_R):
        poisson_solver = self.create_poisson_solver(fine_grid,
                                                    poisson_solver_params)
        return UniformGridPotentialCalculator(grid,
                                              fine_grid,
                                              setups,
                                              xc, poisson_solver,
                                              nct_aR, nct_R)

    def create_hamiltonian_operator(self, grid, blocksize=10):
        return FDHamiltonian(grid, self.stencil, blocksize)


class FDHamiltonian:
    def __init__(self, grid, stencil=3, blocksize=10):
        self.grid = grid
        self.blocksize = blocksize
        self.gd = grid._gd
        self.kin = Laplace(self.gd, -0.5, stencil, grid.dtype)

    def apply(self, vt_sR, psit_nR, out, spin):
        self.kin.apply(psit_nR.data, out.data, psit_nR.desc.phase_factors_cd)
        for p, o in zip(psit_nR.data, out.data):
            o += p * vt_sR.data[spin]
        return out

    def create_preconditioner(self, blocksize):
        from types import SimpleNamespace

        from gpaw.preconditioner import Preconditioner as PC
        pc = PC(self.gd, self.kin, self.grid.dtype, self.blocksize)

        def apply(psit, residuals, out):
            kpt = SimpleNamespace(phase_cd=psit.desc.phase_factors_cd)
            pc(residuals.data, kpt, out=out.data)

        return apply



import ase.io.ulm as ulm
import gpaw
import numpy as np
from ase.io.trajectory import read_atoms, write_atoms
from ase.units import Bohr, Ha
from gpaw.new.builder import DFTComponentsBuilder
from gpaw.new.calculation import DFTCalculation, DFTState
from gpaw.new.density import Density
from gpaw.new.input_parameters import InputParameters
from gpaw.new.potential import Potential
# from gpaw.new.wave_functions import IBZWaveFunctions
from gpaw.utilities import pack, unpack2
from gpaw.core.atom_arrays import AtomArraysLayout


class OldStuff:
    def get_pseudo_wave_function(self, n):
        return self.calculation.ibzwfs[0].wave_functions.data[n]

    def get_atomic_electrostatic_potentials(self):
        _, _, Q_aL = self.calculation.pot_calc.calculate(
            self.calculation.state.density)
        Q_aL = Q_aL.gather()
        return Q_aL.data[::9] * (Ha / (4 * np.pi)**0.5)

    def write(self, filename, mode=''):
        """Write calculator object to a file.

        Parameters
        ----------
        filename
            File to be written
        mode
            Write mode. Use ``mode='all'``
            to include wave functions in the file.
        """
        self.log(f'Writing to {filename} (mode={mode!r})\n')

        write_gpw(filename, self.atoms, self.params,
                  self.calculation, skip_wfs=mode != 'all')


def write_gpw(filename: str,
              atoms,
              params,
              calculation: DFTCalculation,
              skip_wfs: bool = True) -> None:

    world = params.parallel['world']

    if world.rank == 0:
        writer = ulm.Writer(filename, tag='gpaw')
    else:
        writer = ulm.DummyWriter()

    with writer:
        writer.write(version=4,
                     gpaw_version=gpaw.__version__,
                     ha=Ha,
                     bohr=Bohr)

        write_atoms(writer.child('atoms'), atoms)
        writer.child('results').write(**calculation.results)
        writer.child('parameters').write(**params.params)

        density = calculation.state.density
        dms = density.density_matrices.collect()

        N = sum(i1 * (i1 + 1) // 2 for i1, i2 in dms.layout.shapes)
        D = np.zeros((density.ncomponents, N))

        n1 = 0
        for D_iis in dms.values():
            i1 = len(D_iis)
            n2 = n1 + i1 * (i1 + 1) // 2
            for s, D_ii in enumerate(D_iis.T):
                D[s, n1:n2] = pack(D_ii)
            n1 = n2

        writer.child('density').write(
            density=density.nt_s.collect().data * Bohr**-3,
            atomic_density_matrices=D)

        calculation.state.potential.write(writer.child('hamiltonian'))
        calculation.state.ibzwfs.write(writer.child('wave_functions'),
                                       skip_wfs)

    world.barrier()


def read_gpw(filename, log, parallel):
    log(f'Reading from {filename}')

    world = parallel['world']

    reader = ulm.Reader(filename)
    atoms = read_atoms(reader.atoms)

    kwargs = reader.parameters.asdict()
    kwargs['parallel'] = parallel
    params = InputParameters(kwargs)

    builder = DFTComponentsBuilder(atoms, params)
    kpt_band_comm = builder.communicators['D']

    if world.rank == 0:
        nt_sR_array = reader.density.density
        vt_sR_array = reader.hamiltonian.potential
        D_sap_array = reader.density.atomic_density_matrices
        D_saii_array = unpack_d(D_sap_array, builder.setups)
        dH_sap_array = reader.hamiltonian.atomic_hamiltonian_matrices
        dH_saii_array = unpack_d(dH_sap_array, builder.setups)
    else:
        nt_sR_array = None
        vt_sR_array = None
        D_saii_array = None
        dH_saii_array = None

    nt_sR = builder.grid.empty(builder.ncomponents)
    vt_sR = builder.grid.empty(builder.ncomponents)

    atom_array_layout = AtomArraysLayout([(setup.ni, setup.ni)
                                          for setup in builder.setups],
                                         atomdist=builder.atomdist)
    D_asii = atom_array_layout.empty(builder.ncomponents)
    dH_asii = atom_array_layout.empty(builder.ncomponents)

    if kpt_band_comm.rank == 0:
        nt_sR.scatter_from(nt_sR_array)
        vt_sR.scatter_from(vt_sR_array)
        D_asii.scatter_from(D_saii_array)
        dH_asii.scatter_from(dH_saii_array)

    kpt_band_comm.broadcast(nt_sR.data, 0)
    kpt_band_comm.broadcast(vt_sR.data, 0)
    kpt_band_comm.broadcast(D_asii.data, 0)
    kpt_band_comm.broadcast(dH_asii.data, 0)

    density = Density.from_data_and_setups(nt_sR, D_asii,
                                           builder.params.charge,
                                           builder.setups)
    potential = Potential(vt_sR, dH_asii, {})
    ibzwfs = ...

    calculation = DFTCalculation(
        DFTState(ibzwfs, density, potential),
        builder.setups,
        None,
        pot_calc=builder.create_potential_calculator())

    results = reader.results.asdict()
    if results:
        log(f'Read {", ".join(sorted(results))}')

    calculation.results = results
    return calculation, params


def unpack_d(D_sap_array, setups):
    ns = len(D_sap_array)
    naii = sum(setup.ni**2 for setup in setups)
    D_saii_array = np.empty((ns, naii))
    nap1 = 0
    naii1 = 0
    for setup in setups:
        nap2 = nap1 + (setup.ni * (setup.ni + 1)) // 2
        naii2 = naii1 + setup.ni**2
        D_saii_array[:, naii1:naii2] = unpack2(
            D_sap_array[:, nap1:nap2]).ravel()
        nap1 = nap2
        naii1 = naii2
    return D_saii_array


if __name__ == '__main__':
    import sys
    from gpaw.mpi import world
    read_gpw(sys.argv[1], print, {'world': world})



from __future__ import annotations

from typing import Any

import numpy as np
from ase.units import Bohr, Ha
from gpaw.new.builder import DFTComponentsBuilder
from gpaw.new.input_parameters import InputParameters
from gpaw.new.wave_functions import IBZWaveFunctions
from gpaw.new.potential import Potential


class DFTState:
    def __init__(self,
                 ibzwfs: IBZWaveFunctions,
                 density,
                 potential: Potential,
                 vHt_x=None,
                 nct_R=None):
        """State of a Kohn-Sham calculation."""
        self.ibzwfs = ibzwfs
        self.density = density
        self.potential = potential
        self.vHt_x = vHt_x  # initial guess for Hartree potential

    def move(self, fracpos_ac, delta_nct_R):
        self.ibzwfs.move(fracpos_ac)
        self.potential.energies.clear()
        self.density.move(delta_nct_R)


class DFTCalculation:
    def __init__(self,
                 state: DFTState,
                 setups,
                 scf_loop,
                 pot_calc):
        self.state = state
        self.setups = setups
        self.scf_loop = scf_loop
        self.pot_calc = pot_calc

        self.results: dict[str, Any] = {}

    @classmethod
    def from_parameters(cls,
                        atoms,
                        params=None,
                        log=None,
                        builder=None) -> DFTCalculation:

        if isinstance(params, dict):
            params = InputParameters(params)

        builder = builder or DFTComponentsBuilder(atoms, params)

        basis_set = builder.create_basis_set()

        density = builder.density_from_superposition(basis_set)
        density.normalize()

        pot_calc = builder.create_potential_calculator()
        potential, vHt_x, _ = pot_calc.calculate(density)

        if params.random:
            log('Initializing wave functions with random numbers')
            ibzwfs = builder.random_ibz_wave_functions()
        else:
            ibzwfs = builder.lcao_ibz_wave_functions(basis_set, potential)

        write_atoms(atoms, builder.grid, builder.initial_magmoms, log)

        log(ibzwfs.ibz.symmetries)
        log(ibzwfs)

        return cls(DFTState(ibzwfs, density, potential, vHt_x),
                   builder.setups,
                   builder.create_scf_loop(pot_calc),
                   pot_calc)

    def move_atoms(self, atoms, log) -> DFTCalculation:
        self.fracpos_ac = atoms.get_scaled_positions()

        delta_nct_R = self.pot_calc.move(self.fracpos_ac,
                                         self.state.density.ndensities)
        self.state.move(self.fracpos_ac, delta_nct_R)

        magmoms = self.results.get('magmoms')
        write_atoms(atoms,
                    self.state.density.nt_sR.desc,
                    magmoms,
                    log)

        self.results = {}

        return self

    def iconverge(self, log, convergence=None, maxiter=None):
        log(self.scf_loop)
        for ctx in self.scf_loop.iterate(self.state,
                                         convergence,
                                         maxiter,
                                         log=log):
            yield ctx

    def converge(self,
                 log,
                 convergence=None,
                 maxiter=None,
                 steps=99999999999999999):
        """Converge to self-consistent solution of KS-equation."""
        for step, _ in enumerate(self.iconverge(log, convergence, maxiter),
                                 start=1):
            if step == steps:
                break
        else:  # no break
            log(f'Converged in {step} steps')

    def energies(self, log):
        energies1 = self.state.potential.energies.copy()
        energies2 = self.state.ibzwfs.energies
        energies1['kinetic'] += energies2['band']
        energies1['entropy'] = energies2['entropy']
        free_energy = sum(energies1.values())
        extrapolated_energy = free_energy + energies2['extrapolation']

        log('\nEnergies (eV):')
        for name, e in energies1.items():
            log(f'    {name + ":":10}   {e * Ha:14.6f}')
        log(f'    Total:       {free_energy * Ha:14.6f}')
        log(f'    Extrapolated:{extrapolated_energy * Ha:14.6f}')

        self.results['free_energy'] = free_energy
        self.results['energy'] = extrapolated_energy

    def dipole(self, log):
        dipole_v = self.density.calculate_dipole_moment() * Bohr
        self.log('Dipole moment: ({:.6f}, {:.6f}, {:.6f}) |e|*Ang\n'
                 .format(*dipole_v))
        self.results['dipole'] = dipole_v

    def magmoms(self, log):
        mm_v, mm_av = self.state.density.calculate_magnetic_moments()
        self.results['magmom'] = mm_v[2]
        self.results['magmoms'] = mm_av[:, 2].copy()

        if self.state.density.ncomponents > 1:
            x, y, z = mm_v
            log(f'Total magnetic moment: ({x:.6f}, {y:.6f}, {z:.6f})')
            log('Local magnetic moments:')
            for a, (setup, m_v) in enumerate(zip(self.setups, mm_av)):
                x, y, z = m_v
                log(f'{a:4} {setup.symbol:2} ({x:9.6f}, {y:9.6f}, {z:9.6f})')
            log()

    def forces(self, log):
        """Return atomic force contributions."""
        xc = self.pot_calc.xc
        assert not xc.no_forces
        assert not hasattr(xc.xc, 'setup_force_corrections')

        # Force from projector functions (and basis set):
        F_av = self.state.ibzwfs.forces(self.state.potential.dH_asii)

        pot_calc = self.pot_calc
        Fcc_aLv, Fnct_av, Fvbar_av = pot_calc.force_contributions(
            self.state)

        # Force from compensation charges:
        ccc_aL = \
            self.state.density.calculate_compensation_charge_coefficients()
        for a, dF_Lv in Fcc_aLv.items():
            F_av[a] += ccc_aL[a] @ dF_Lv

        # Force from smooth core charge:
        for a, dF_v in Fnct_av.items():
            F_av[a] += dF_v[0]

        # Force from zero potential:
        for a, dF_v in Fvbar_av.items():
            F_av[a] += dF_v[0]

        domain_comm = ccc_aL.layout.atomdist.comm
        domain_comm.sum(F_av)

        F_av = self.state.ibzwfs.ibz.symmetries.symmetrize_forces(F_av)

        log('\nForces [eV/Ang]:')
        c = Ha / Bohr
        for a, setup in enumerate(self.setups):
            x, y, z = F_av[a] * c
            log(f'{a:4} {setup.symbol:2} {x:10.3f} {y:10.3f} {z:10.3f}')

        self.results['forces'] = F_av

    def write_converged(self, log):
        self.state.ibzwfs.write_summary(log)


def write_atoms(atoms, grid, magmoms, log):
    from gpaw.output import print_cell, print_positions
    if magmoms is None:
        magmoms = np.zeros((len(atoms), 3))
    elif magmoms.ndim == 1:
        m1 = magmoms
        magmoms = np.zeros((len(atoms), 3))
        magmoms[:, 2] = m1
    print_positions(atoms, log, magmoms)
    print_cell(grid._gd, atoms.pbc, log)



from __future__ import annotations

from functools import partial
from typing import Sequence

import numpy as np
from ase.units import Ha
from gpaw.core.arrays import DistributedArrays as DA
from gpaw.core.atom_arrays import AtomArrays
from gpaw.mpi import MPIComm
from gpaw.new.brillouin import IBZ
from gpaw.setup import Setups
from gpaw.typing import Array1D, Array2D, ArrayND
from gpaw.utilities.debug import frozen


@frozen
class IBZWaveFunctions:
    def __init__(self,
                 ibz: IBZ,
                 rank_k: Sequence[int],
                 kpt_comm: MPIComm,
                 wfs_qs: list[list[WaveFunctions]],
                 nelectrons: float,
                 spin_degeneracy: int = 2):
        """Collection of wave function objects for k-points in the IBZ."""
        self.ibz = ibz
        self.rank_k = rank_k
        self.kpt_comm = kpt_comm
        self.wfs_qs = wfs_qs
        self.nelectrons = nelectrons
        self.fermi_levels = None
        self.collinear = False
        self.spin_degeneracy = spin_degeneracy

        self.band_comm = wfs_qs[0][0].psit_nX.comm
        self.domain_comm = wfs_qs[0][0].psit_nX.desc.comm

        self.nbands = wfs_qs[0][0].psit_nX.dims[0]

        self.q_k = {}  # ibz index to local index
        q = 0
        for k, rank in enumerate(rank_k):
            if rank == kpt_comm.rank:
                self.q_k[k] = q
                q += 1
        self.energies: dict[str, float] = {}

    def __str__(self):
        return str(self.ibz)

    def __iter__(self):
        for wfs_s in self.wfs_qs:
            yield from wfs_s

    @classmethod
    def from_random_numbers(cls,
                            ibz,
                            band_comm,
                            kpt_comm,
                            desc,
                            setups,
                            fracpos_ac,
                            nbands: int,
                            nelectrons: float,
                            dtype=None) -> IBZWaveFunctions:
        rank_k = ibz.ranks(kpt_comm)

        wfs_q = []
        for kpt_c, weight, rank in zip(ibz.kpt_kc, ibz.weight_k, rank_k):
            if rank != kpt_comm.rank:
                continue
            desck = desc.new(kpt=kpt_c, dtype=dtype)
            wfs = WaveFunctions.from_random_numbers(desck, weight,
                                                    nbands, band_comm,
                                                    setups,
                                                    fracpos_ac)
            wfs_q.append(wfs)

        return cls(ibz, rank_k, kpt_comm, wfs_q, nelectrons)

    def move(self, fracpos_ac):
        self.ibz.symmetries.check_positions(fracpos_ac)
        self.energies.clear()
        for wfs in self:
            wfs._P_ain = None
            wfs.orthonormalized = False
            wfs.pt_aiX.move(fracpos_ac)
            wfs._eig_n = None
            wfs._occ_n = None

    def orthonormalize(self, work_array_nX: ArrayND = None):
        for wfs in self:
            wfs.orthonormalize(work_array_nX)

    def calculate_occs(self, occ_calc, fixed_fermi_level=False):
        degeneracy = self.spin_degeneracy

        # u index is q and s combined
        occ_un, fermi_levels, e_entropy = occ_calc.calculate(
            nelectrons=self.nelectrons / degeneracy,
            eigenvalues=[wfs.eig_n * Ha for wfs in self],
            weights=[wfs.weight for wfs in self],
            fermi_levels_guess=(None
                                if self.fermi_levels is None else
                                self.fermi_levels * Ha))

        if not fixed_fermi_level or self.fermi_levels is None:
            self.fermi_levels = np.array(fermi_levels) / Ha

        for occ_n, wfs in zip(occ_un, self):
            wfs._occ_n = occ_n

        e_entropy *= degeneracy / Ha
        e_band = 0.0
        for wfs in self:
            e_band += wfs.occ_n @ wfs.eig_n * wfs.weight * degeneracy
        e_band = self.kpt_comm.sum(e_band)
        self.energies = {
            'band': e_band,
            'entropy': e_entropy,
            'extrapolation': e_entropy * occ_calc.extrapolate_factor}

    def add_to_density(self, nt_sR, D_asii) -> None:
        for wfs in self:
            wfs.add_to_density(nt_sR, D_asii)
        self.kpt_comm.sum(nt_sR.data)
        self.kpt_comm.sum(D_asii.data)

    def get_eigs_and_occs(self, k=0, s=0):
        if self.domain_comm.rank == 0 and self.band_comm.rank == 0:
            rank = self.rank_k[k]
            if rank == self.kpt_comm.rank:
                wfs = self.wfs_qs[self.q_k[k]][s]
                if rank == 0:
                    return wfs._eig_n, wfs._occ_n
                self.kpt_comm.send(wfs._eig_n, 0)
                self.kpt_comm.send(wfs._occ_n, 0)
            elif self.kpt_comm.rank == 0:
                eig_n = np.empty(self.nbands)
                occ_n = np.empty(self.nbands)
                self.kpt_comm.receive(eig_n, rank)
                self.kpt_comm.receive(occ_n, rank)
                return eig_n, occ_n
        return np.zeros(0), np.zeros(0)

    def forces(self, dH_asii: AtomArrays):
        F_av = np.zeros((dH_asii.natoms, 3))
        for wfs in self:
            wfs.force_contribution(dH_asii, F_av)
        self.kpt_comm.sum(F_av)
        return F_av

    def write(self, writer, skip_wfs):
        writer.write(fermi_levels=self.fermi_levels)

    def write_summary(self, log):
        fl = self.fermi_levels * Ha
        assert len(fl) == 1
        log(f'\nFermi level: {fl[0]:.3f}')

        ibz = self.ibz

        for k, (x, y, z) in enumerate(ibz.kpt_kc):
            log(f'\nkpt = [{x:.3f}, {y:.3f}, {z:.3f}], '
                f'weight = {ibz.weight_k[k]:.3f}:')

            if self.spin_degeneracy == 2:
                log('  Band      eig [eV]   occ [0-2]')
                eigs, occs = self.get_eigs_and_occs(k)
                for n, (e, f) in enumerate(zip(eigs * Ha, occs)):
                    log(f'  {n:4} {e:13.3f}   {2 * f:9.3f}')
            else:
                log('  Band      eig [eV]   occ [0-1]'
                    '      eig [eV]   occ [0-1]')
                eigs1, occs1 = self.get_eigs_and_occs(k, 0)
                eigs2, occs2 = self.get_eigs_and_occs(k, 1)
                for n, (e1, f1, e2, f2) in enumerate(zip(eigs1 * Ha,
                                                         occs1,
                                                         eigs2 * Ha,
                                                         occs2)):
                    log(f'  {n:4} {e1:13.3f}   {f1:9.3f}'
                        f'    {e2:10.3f}   {f2:9.3f}')
            if k == 3:
                break


@frozen
class WaveFunctions:
    def __init__(self,
                 psit_nX: DA,
                 spin: int | None,
                 setups: Setups,
                 fracpos_ac: Array2D,
                 weight: float = 1.0,
                 spin_degeneracy: int = 2):
        self.psit_nX = psit_nX
        self.spin = spin
        self.setups = setups
        self.weight = weight
        self.spin_degeneracy = spin_degeneracy

        self._P_ain = None
        self.pt_aiX = setups.create_projectors(self.psit_nX.desc,
                                               fracpos_ac)
        self.orthonormalized = False

        self._eig_n: Array1D | None = None
        self._occ_n: Array1D | None = None

    @property
    def eig_n(self) -> Array1D:
        if self._eig_n is None:
            raise ValueError
        return self._eig_n

    @property
    def occ_n(self) -> Array1D:
        if self._occ_n is None:
            raise ValueError
        return self._occ_n

    @property
    def myeig_n(self):
        assert self.psit_nX.comm.size == 1
        return self.eig_n

    @property
    def myocc_n(self):
        assert self.psit_nX.comm.size == 1
        return self.occ_n

    @property
    def P_ain(self):
        if self._P_ain is None:
            self._P_ain = self.pt_aiX.empty(self.psit_nX.dims,
                                            self.psit_nX.comm,
                                            transposed=True)
            self.pt_aiX.integrate(self.psit_nX, self._P_ain)
        return self._P_ain

    @classmethod
    def from_random_numbers(cls, desc, weight, nbands, band_comm, setups,
                            fracpos_ac):
        wfs = desc.random(nbands, band_comm)
        return cls(wfs, 0, setups, fracpos_ac, weight)

    def add_to_density(self,
                       nt_sR,
                       D_asii: AtomArrays) -> None:
        occ_n = self.weight * self.spin_degeneracy * self.myocc_n
        self.psit_nX.abs_square(weights=occ_n, out=nt_sR[self.spin])

        for D_sii, P_in in zip(D_asii.values(), self.P_ain.values()):
            D_sii[self.spin] += np.einsum('in, n, jn -> ij',
                                          P_in.conj(), occ_n, P_in).real

    def orthonormalize(self, work_array_nX: ArrayND = None):
        if self.orthonormalized:
            return
        psit_nX = self.psit_nX
        domain_comm = psit_nX.desc.comm

        P_ain = self.P_ain

        P2_ain = P_ain.new()
        psit2_nX = psit_nX.new(data=work_array_nX)

        dS = self.setups.overlap_correction

        S = psit_nX.matrix_elements(psit_nX, domain_sum=False)
        dS(P_ain, out=P2_ain)
        P_ain.matrix.multiply(P2_ain, opa='C', symmetric=True, out=S, beta=1.0)
        domain_comm.sum(S.data, 0)

        if domain_comm.rank == 0:
            S.invcholesky()
        # S now contains the inverse of the Cholesky factorization
        domain_comm.broadcast(S.data, 0)
        # cc ??????

        S.multiply(psit_nX, out=psit2_nX)
        P_ain.matrix.multiply(S, opb='T', out=P2_ain)
        psit_nX.data[:] = psit2_nX.data
        P_ain.data[:] = P2_ain.data

        self.orthonormalized = True

    def subspace_diagonalize(self,
                             Ht,
                             dH,
                             work_array: ArrayND = None,
                             Htpsit_nX=None,
                             scalapack_parameters=(None, 1, 1, -1)):
        """

        Ht(in, out)::

           ~   ^   ~
           H = T + v

        dH::

            ~  ~    a    ~  ~
          <psi|p> dH    <p|psi>
              m i   ij    j   n
        """
        self.orthonormalize(work_array)
        psit_nX = self.psit_nX
        P_ain = self.P_ain
        psit2_nX = psit_nX.new(data=work_array)
        P2_ain = P_ain.new()
        domain_comm = psit_nX.desc.comm

        Ht = partial(Ht, out=psit2_nX, spin=self.spin)

        H = psit_nX.matrix_elements(psit_nX, function=Ht, domain_sum=False)
        dH(P_ain, out=P2_ain, spin=self.spin)
        P_ain.matrix.multiply(P2_ain, opa='C', symmetric=True,
                              out=H, beta=1.0)
        domain_comm.sum(H.data, 0)

        if domain_comm.rank == 0:
            slcomm, r, c, b = scalapack_parameters
            if r == c == 1:
                slcomm = None
            self._eig_n = H.eigh(scalapack=(slcomm, r, c, b))
            # H.data[n, :] now contains the n'th eigenvector and eps_n[n]
            # the n'th eigenvalue
        else:
            self._eig_n = np.empty(psit_nX.dims)

        domain_comm.broadcast(H.data, 0)
        domain_comm.broadcast(self._eig_n, 0)
        if Htpsit_nX is not None:
            H.multiply(psit2_nX, out=Htpsit_nX)

        H.multiply(psit_nX, out=psit2_nX)
        psit_nX.data[:] = psit2_nX.data
        P_ain.matrix.multiply(H, opb='T', out=P2_ain)
        P_ain.data[:] = P2_ain.data

    def force_contribution(self, dH_asii: AtomArrays, F_av: Array2D):
        F_ainv = self.pt_aiX.derivative(self.psit_nX)
        myocc_n = self.weight * self.spin_degeneracy * self.myocc_n
        for a, F_inv in F_ainv.items():
            F_inv = F_inv.conj()
            F_inv *= myocc_n[:, np.newaxis]
            dH_ii = dH_asii[a][self.spin]
            P_in = self.P_ain[a]
            F_vii = np.einsum('inv, jn, jk -> vik', F_inv, P_in, dH_ii)
            F_inv *= self.myeig_n[:, np.newaxis]
            dO_ii = self.setups[a].dO_ii
            F_vii -= np.einsum('inv, jn, jk -> vik', F_inv, P_in, dO_ii)
            F_av[a] += 2 * F_vii.real.trace(0, 1, 2)



from types import SimpleNamespace

import numpy as np
from gpaw.band_descriptor import BandDescriptor
from gpaw.core import PlaneWaves
from gpaw.kohnsham_layouts import get_KohnSham_layouts
from gpaw.lcao.eigensolver import DirectLCAO
from gpaw.new.potential import Potential
from gpaw.new.wave_functions import IBZWaveFunctions, WaveFunctions
from gpaw.utilities import pack2
from gpaw.utilities.partition import AtomPartition
from gpaw.utilities.timing import nulltimer
from gpaw.wavefunctions.lcao import LCAOWaveFunctions


def create_lcao_ibz_wave_functions(setups,
                                   communicators,
                                   nbands,
                                   ncomponents,
                                   nelectrons,
                                   fracpos_ac,
                                   dtype,
                                   grid,
                                   wf_desc,
                                   ibz,
                                   sl_lcao,
                                   basis_set,
                                   potential: Potential) -> IBZWaveFunctions:
    band_comm = communicators['b']
    domain_comm = communicators['d']
    domainband_comm = communicators['K']
    kptband_comm = communicators['D']
    kpt_comm = communicators['k']
    world = communicators['w']
    lcaonbands = min(nbands, setups.nao)
    gd = grid._gd

    lcaobd = BandDescriptor(lcaonbands, band_comm)
    lcaoksl = get_KohnSham_layouts(sl_lcao, 'lcao',
                                   gd, lcaobd, domainband_comm,
                                   dtype, nao=setups.nao)

    atom_partition = AtomPartition(
        domain_comm,
        np.array([sphere.rank for sphere in basis_set.sphere_a]))

    lcaowfs = LCAOWaveFunctions(lcaoksl, gd, nelectrons,
                                setups, lcaobd, dtype,
                                world, basis_set.kd, kptband_comm,
                                nulltimer)
    lcaowfs.basis_functions = basis_set
    lcaowfs.set_positions(fracpos_ac, atom_partition)

    if ncomponents != 4:
        eigensolver = DirectLCAO()
    else:
        from gpaw.xc.noncollinear import NonCollinearLCAOEigensolver
        eigensolver = NonCollinearLCAOEigensolver()

    eigensolver.initialize(gd, dtype, setups.nao, lcaoksl)

    dH_asp = setups.empty_atomic_matrix(ncomponents, atom_partition,
                                        dtype=dtype)
    for a, dH_sii in potential.dH_asii.items():
        dH_asp[a][:] = [pack2(dH_ii) for dH_ii in dH_sii]
    ham = SimpleNamespace(vt_sG=potential.vt_sR.data,
                          dH_asp=dH_asp)
    eigensolver.iterate(ham, lcaowfs)

    rank_k = ibz.ranks(kpt_comm)

    nspins = ncomponents % 3
    u = 0
    wfs_qs = []
    for kpt_c, weight, rank in zip(ibz.kpt_kc, ibz.weight_k, rank_k):
        if rank != kpt_comm.rank:
            continue
        gridk = grid.new(kpt=kpt_c, dtype=dtype)
        wfs_s = []
        for s in range(nspins):
            lcaokpt = lcaowfs.kpt_u[u]
            assert (ibz.kpt_kc[lcaokpt.k] == kpt_c).all(), (ibz.kpt_kc,
                                                            lcaokpt.k,
                                                            kpt_c)
            assert lcaokpt.s == s
            psit_nR = gridk.zeros(nbands, band_comm)
            mynbands = len(lcaokpt.C_nM)
            basis_set.lcao_to_grid(lcaokpt.C_nM,
                                   psit_nR.data[:mynbands], lcaokpt.q)
            if isinstance(wf_desc, PlaneWaves):
                pw = wf_desc.new(kpt=kpt_c)
                psit_nG = pw.empty(nbands, band_comm)
                for psit_R, psit_G in zip(psit_nR, psit_nG):
                    psit_R.fft(out=psit_G)
                psit_nX = psit_nG
            else:
                psit_nX = psit_nR
            wfs_s.append(WaveFunctions(psit_nX, s, setups,
                                       fracpos_ac, weight,
                                       spin_degeneracy=2 // nspins))
            assert mynbands == nbands
            u += 1
        wfs_qs.append(wfs_s)

    ibzwfs = IBZWaveFunctions(ibz, rank_k, kpt_comm, wfs_qs, nelectrons,
                              2 // nspins)
    return ibzwfs



from ase import Atoms
from ase.units import Ha

from gpaw import GPAW
from gpaw.test import equal
from gpaw.poisson import NoInteractionPoissonSolver
from gpaw.external import ExternalPotential, known_potentials


class HarmonicPotential(ExternalPotential):
    def calculate_potential(self, gd):
        a = gd.cell_cv[0, 0]
        r_vg = gd.get_grid_point_coordinates()
        self.vext_g = 0.5 * ((r_vg - a / 2)**2).sum(0)

    def todict(self):
        return {'name': 'HarmonicPotential'}


def test_ext_potential_harmonic(in_tmp_dir):
    """Test againts analytic result (no xc, no Coulomb)."""
    a = 4.0
    x = Atoms(cell=(a, a, a))  # no atoms

    calc = GPAW(charge=-8,
                nbands=4,
                h=0.2,
                xc={'name': 'null'},
                external=HarmonicPotential(),
                poissonsolver=NoInteractionPoissonSolver(),
                eigensolver='cg')

    x.calc = calc
    x.get_potential_energy()

    eigs = calc.get_eigenvalues()
    equal(eigs[0], 1.5 * Ha, 0.002)
    equal(abs(eigs[1:] - 2.5 * Ha).max(), 0, 0.003)

    # Check write + read:
    calc.write('harmonic.gpw')
    known_potentials['HarmonicPotential'] = HarmonicPotential
    GPAW('harmonic.gpw')



from __future__ import annotations
from pathlib import Path

from typing import Any, IO, Sequence

import numpy as np
from gpaw.mpi import world

parameter_functions = {}

"""
background_charge
external
reuse_wfs_method
"""


def input_parameter(func):
    parameter_functions[func.__name__] = func


def update_dict(default, value) -> dict[str, Any]:
    dct = default.copy()
    if value is not None:
        assert value.keys() <= default.keys(), (value, default)
        dct.update(value)
    return dct


class InputParameters:
    h: float | None
    parallel: dict[str, Any]
    txt: str | Path | IO[str] | None
    mode: dict[str, Any]
    xc: dict[str, Any]
    symmetry: dict[str, Any]
    kpts: dict[str, Any]
    setups: Any
    basis: Any
    magmoms: Any
    gpts: None | Sequence[int]
    charge: float
    nbands: None | int | float
    spinpol: bool

    def __init__(self, params: dict[str, Any]):
        """Accuracy of the self-consistency cycle."""
        self.params = params

        for key in params:
            if key not in parameter_functions:
                raise ValueError(
                    f'Unknown parameter {key!r}.  Must be one of: ' +
                    ', '.join(parameter_functions))
        for key, func in parameter_functions.items():
            if key in params:
                value = func(params[key])
            else:
                value = func()
            self.__dict__[key] = value

    def __repr__(self) -> str:
        p = ', '.join(f'{key}={value!r}'
                      for key, value in self.params.items())
        return f'InputParameters({p})'


@input_parameter
def occupations(value=None):
    return value


@input_parameter
def poissonsolver(value=None):
    """Poisson solver."""
    return value or {}


@input_parameter
def parallel(value: dict[str, Any] = None) -> dict[str, Any]:
    dct = update_dict({'kpt': None,
                       'domain': None,
                       'band': None,
                       'order': 'kdb',
                       'stridebands': False,
                       'augment_grids': False,
                       'sl_auto': False,
                       'sl_default': None,
                       'sl_diagonalize': None,
                       'sl_inverse_cholesky': None,
                       'sl_lcao': None,
                       'sl_lrtddft': None,
                       'use_elpa': False,
                       'elpasolver': '2stage',
                       'buffer_size': None,
                       'world': None},
                      value)
    dct['world'] = dct['world'] or world
    return dct


@input_parameter
def eigensolver(value=None):
    """Eigensolver."""
    return value or {'converge': 'occupied'}


@input_parameter
def charge(value=0.0):
    return value


@input_parameter
def mixer(value=None):
    return value or {}


@input_parameter
def hund(value=False):
    """Using Hund's rule for guessing initial magnetic moments."""
    return value


@input_parameter
def xc(value='LDA'):
    """Exchange-Correlation functional."""
    if isinstance(value, str):
        return {'name': value}


@input_parameter
def mode(value='fd'):
    return {'name': value} if isinstance(value, str) else value


@input_parameter
def setups(value='paw'):
    """PAW datasets or pseudopotentials."""
    return value if isinstance(value, dict) else {None: value}


@input_parameter
def symmetry(value='undefined'):
    """Use of symmetry."""
    if value == 'undefined':
        value = {}
    elif value in {None, 'off'}:
        value = {'point_group': False, 'time_reversal': False}
    return value


@input_parameter
def basis(value=None):
    """Atomic basis set."""
    return value or {}


@input_parameter
def magmoms(value=None):
    return value


@input_parameter
def kpts(value=None) -> dict[str, Any]:
    """Brillouin-zone sampling."""
    if value is None:
        value = {'size': (1, 1, 1)}
    elif not isinstance(value, dict):
        if len(value) == 3 and isinstance(value[0], int):
            value = {'size': value}
        else:
            value = {'points': np.array(value)}
    return value


@input_parameter
def maxiter(value=333):
    """Maximum number of SCF-iterations."""
    return value


@input_parameter
def h(value=None):
    """Grid spacing."""
    return value


@input_parameter
def txt(value: str | Path | IO[str] | None = '?'
        ) -> str | Path | IO[str] | None:
    """Log file."""
    return value


@input_parameter
def random(value=False):
    return value


@input_parameter
def spinpol(value=False):
    return value


@input_parameter
def gpts(value=None):
    """Number of grid points."""
    return value


@input_parameter
def nbands(value: str | int | None = None) -> int | float | None:
    """Number of electronic bands."""
    if isinstance(value, int) or value is None:
        return value
    if nbands[-1] == '%':
        return float(value[:-1]) / 100
    raise ValueError('Integer expected: Only use a string '
                     'if giving a percentage of occupied bands')


@input_parameter
def soc(value=False):
    return value


@input_parameter
def convergence(value=None):
    return update_dict({'energy': 0.0005,  # eV / electron
                        'density': 1.0e-4,  # electrons / electron
                        'eigenstates': 4.0e-8,  # eV^2 / electron
                        'forces': np.inf},
                       value)



from __future__ import annotations
from gpaw.core.arrays import DistributedArrays
from gpaw.core.atom_arrays import AtomArrays


class Potential:
    def __init__(self,
                 vt_sR: DistributedArrays,
                 dH_asii: AtomArrays,
                 energies: dict[str, float]):
        self.vt_sR = vt_sR
        self.dH_asii = dH_asii
        self.energies = energies

    def dH(self, P_ain, out, spin):
        for a, I1, I2 in P_ain.layout.myindices:
            dH_ii = self.dH_asii[a][spin]
            out.data[I1:I2] = dH_ii @ P_ain.data[I1:I2]
        return out

    def write(self, writer):
        writer.write(vt=self.vt.collect().data)



from gpaw.xc.hybrid import HybridXCBase


class XCFunctional:
    def __init__(self, xc):
        self.xc = xc
        self.setup_name = xc.get_setup_name()
        self.name = xc.name
        self.no_forces = (isinstance(xc, HybridXCBase) or
                          self.name.startswith('GLLB'))

    def calculate(self, density, out) -> float:
        return self.xc.calculate(density.desc._gd, density.data, out.data)

    def calculate_paw_correction(self, setup, d, h):
        return self.xc.calculate_paw_correction(setup, d, h)



import os
from contextlib import contextmanager
from pathlib import Path

import numpy as np
import pytest
from _pytest.tmpdir import _mk_tmp
from ase import Atoms
from ase.build import bulk
from ase.io import read

from gpaw import GPAW, PW, Davidson, FermiDirac
from gpaw.cli.info import info
from gpaw.mpi import broadcast, world
from gpaw.utilities import devnull


@contextmanager
def execute_in_tmp_path(request, tmp_path_factory):
    if world.rank == 0:
        # Obtain basename as
        # * request.function.__name__  for function fixture
        # * request.module.__name__    for module fixture
        basename = getattr(request, request.scope).__name__
        path = tmp_path_factory.mktemp(basename)
    else:
        path = None
    path = broadcast(path)
    cwd = os.getcwd()
    os.chdir(path)
    try:
        yield path
    finally:
        os.chdir(cwd)


@pytest.fixture(scope='function')
def in_tmp_dir(request, tmp_path_factory):
    """Run test function in a temporary directory."""
    with execute_in_tmp_path(request, tmp_path_factory) as path:
        yield path


@pytest.fixture(scope='module')
def module_tmp_path(request, tmp_path_factory):
    """Run test module in a temporary directory."""
    with execute_in_tmp_path(request, tmp_path_factory) as path:
        yield path


@pytest.fixture(scope='session')
def gpw_files(request, tmp_path_factory):
    """Reuse gpw-files.

    Returns a dict mapping names to paths to gpw-files.  If you
    want to reuse gpw-files from an earlier pytest session then set the
    ``$GPW_TEST_FILES`` environment variable and the files will be written
    to that folder.

    Example::

        def test_something(gpw_files):
            calc = GPAW(gpw_files['h2_lcao_wfs'])
            ...

    Possible systems are:

    * Bulk BCC-Li with 3x3x3 k-points: ``bcc_li_pw``, ``bcc_li_fd``,
      ``bcc_li_lcao``.

    * O2 molecule: ``o2_pw``.

    * H2 molecule: ``h2_pw``, ``h2_fd``, ``h2_lcao``.

    * H2 molecule (not centered): ``h2_pw_0``.

    * Spin-polarized H atom: ``h_pw``.

    * Polyethylene chain.  One unit, 3 k-points, no symmetry:
      ``c2h4_pw_nosym``.  Three units: ``c6h12_pw``.

    * Bulk TiO2 with 4x4x4 k-points: ``ti2o4_pw`` and ``ti2o4_pw_nosym``.

    Files with wave functions are also availabel (add ``_wfs`` to the names).
    """
    path = os.environ.get('GPW_TEST_FILES')
    if path is None:
        if world.rank == 0:
            path = _mk_tmp(request, tmp_path_factory)
        else:
            path = None
        path = broadcast(path)
    return GPWFiles(Path(path))


class GPWFiles:
    """Create gpw-files."""
    def __init__(self, path: Path):
        self.path = path
        self.gpw_files = {}
        for file in path.glob('*.gpw'):
            self.gpw_files[file.name[:-4]] = file

    def __getitem__(self, name: str) -> Path:
        if name not in self.gpw_files:
            rawname, _, _ = name.partition('_wfs')
            calc = getattr(self, rawname)()
            path = self.path / (rawname + '.gpw')
            calc.write(path)
            self.gpw_files[rawname] = path
            path = self.path / (rawname + '_wfs.gpw')
            calc.write(path, mode='all')
            self.gpw_files[rawname + '_wfs'] = path
        return self.gpw_files[name]

    def bcc_li_pw(self):
        return self.bcc_li({'name': 'pw', 'ecut': 200})

    def bcc_li_fd(self):
        return self.bcc_li({'name': 'fd'})

    def bcc_li_lcao(self):
        return self.bcc_li({'name': 'lcao'})

    def bcc_li(self, mode):
        li = bulk('Li', 'bcc', 3.49)
        li.calc = GPAW(mode=mode,
                       kpts=(3, 3, 3),
                       txt=self.path / f'bcc_li_{mode["name"]}.txt')
        li.get_potential_energy()
        return li.calc

    def h2_pw(self):
        return self.h2({'name': 'pw', 'ecut': 200})

    def h2_fd(self):
        return self.h2({'name': 'fd'})

    def h2_lcao(self):
        return self.h2({'name': 'lcao'})

    def h2(self, mode):
        h2 = Atoms('H2', positions=[[0, 0, 0], [0.74, 0, 0]])
        h2.center(vacuum=2.5)
        h2.calc = GPAW(mode=mode,
                       txt=self.path / f'h2_{mode["name"]}.txt')
        h2.get_potential_energy()
        return h2.calc

    def h2_pw_0(self):
        h2 = Atoms('H2',
                   positions=[[-0.37, 0, 0], [0.37, 0, 0]],
                   cell=[5.74, 5, 5])
        h2.calc = GPAW(mode={'name': 'pw', 'ecut': 200},
                       txt=self.path / 'h2_pw_0.txt')
        h2.get_potential_energy()
        return h2.calc

    def h_pw(self):
        h = Atoms('H', magmoms=[1])
        h.center(vacuum=4.0)
        h.calc = GPAW(mode={'name': 'pw', 'ecut': 500},
                      txt=self.path / 'h_pw.txt')
        h.get_potential_energy()
        return h.calc

    def o2_pw(self):
        d = 1.1
        h = Atoms('O2', positions=[[0, 0, 0], [d, 0, 0]], magmoms=[1, 1])
        h.center(vacuum=4.0)
        h.calc = GPAW(mode={'name': 'pw', 'ecut': 800},
                      txt=self.path / 'o2_pw.txt')
        h.get_potential_energy()
        return h.calc

    def c2h4_pw_nosym(self):
        d = 1.54
        h = 1.1
        x = d * (2 / 3)**0.5
        z = d / 3**0.5
        pe = Atoms('C2H4',
                   positions=[[0, 0, 0],
                              [x, 0, z],
                              [0, -h * (2 / 3)**0.5, -h / 3**0.5],
                              [0, h * (2 / 3)**0.5, -h / 3**0.5],
                              [x, -h * (2 / 3)**0.5, z + h / 3**0.5],
                              [x, h * (2 / 3)**0.5, z + h / 3**0.5]],
                   cell=[2 * x, 0, 0],
                   pbc=(1, 0, 0))
        pe.center(vacuum=2.0, axis=(1, 2))
        pe.calc = GPAW(mode='pw',
                       kpts=(3, 1, 1),
                       symmetry='off',
                       txt=self.path / 'c2h4_pw_nosym.txt')
        pe.get_potential_energy()
        return pe.calc

    def c6h12_pw(self):
        pe = read(self['c2h4_pw_nosym'])
        pe = pe.repeat((3, 1, 1))
        pe.calc = GPAW(mode='pw', txt=self.path / 'c6h12_pw.txt')
        pe.get_potential_energy()
        return pe.calc

    def h2o_lcao(self):
        from ase.build import molecule
        atoms = molecule('H2O', cell=[8, 8, 8], pbc=1)
        atoms.center()
        atoms.calc = GPAW(mode='lcao', txt=self.path / 'h2o.txt')
        atoms.get_potential_energy()
        return atoms.calc

    def ti2o4(self, symmetry):
        pwcutoff = 400.0
        k = 4
        a = 4.59
        c = 2.96
        u = 0.305

        rutile_cell = [[a, 0, 0],
                       [0, a, 0],
                       [0, 0, c]]

        TiO2_basis = np.array([[0.0, 0.0, 0.0],
                               [0.5, 0.5, 0.5],
                               [u, u, 0.0],
                               [-u, -u, 0.0],
                               [0.5 + u, 0.5 - u, 0.5],
                               [0.5 - u, 0.5 + u, 0.5]])

        bulk_crystal = Atoms(symbols='Ti2O4',
                             scaled_positions=TiO2_basis,
                             cell=rutile_cell,
                             pbc=(1, 1, 1))

        tag = '_nosym' if symmetry == 'off' else ''
        bulk_calc = GPAW(mode=PW(pwcutoff),
                         nbands=42,
                         eigensolver=Davidson(1),
                         kpts={'size': (k, k, k), 'gamma': True},
                         xc='PBE',
                         occupations=FermiDirac(0.00001),
                         parallel={'band': 1},
                         symmetry=symmetry,
                         txt=self.path / f'ti2o4_pw{tag}.txt')

        bulk_crystal.calc = bulk_calc
        bulk_crystal.get_potential_energy()
        return bulk_calc

    def ti2o4_pw(self):
        return self.ti2o4({})

    def ti2o4_pw_nosym(self):
        return self.ti2o4('off')


class GPAWPlugin:
    def __init__(self):
        if world.rank == -1:
            print()
            info()

    def pytest_terminal_summary(self, terminalreporter, exitstatus, config):
        from gpaw.mpi import size
        terminalreporter.section('GPAW-MPI stuff')
        terminalreporter.write(f'size: {size}\n')


def pytest_configure(config):
    if world.rank != 0:
        try:
            tw = config.get_terminal_writer()
        except AttributeError:
            pass
        else:
            tw._file = devnull
    config.pluginmanager.register(GPAWPlugin(), 'pytest_gpaw')
    for line in ['soc: Spin-orbit coupling',
                 'slow: slow test',
                 'fast: fast test',
                 'ci: test for CI',
                 'libxc: LibXC requirered',
                 'mgga: MGGA test',
                 'dscf: Delta-SCF',
                 'mom: MOM',
                 'gllb: GLLBSC tests',
                 'elph: Electron-phonon',
                 'intel: fails on INTEL toolchain',
                 'serial: run in serial only']:
        config.addinivalue_line('markers', line)


def pytest_runtest_setup(item):
    """Skip some tests.

    If:

    * they depend on libxc and GPAW is not compiled with libxc
    * they are before $PYTEST_START_AFTER
    """
    from gpaw import libraries

    if world.size > 1:
        for mark in item.iter_markers():
            if mark.name == 'serial':
                pytest.skip('Only run in serial')

    if item.location[0] <= os.environ.get('PYTEST_START_AFTER', ''):
        pytest.skip('Not after $PYTEST_START_AFTER')
        return

    if libraries['libxc']:
        return

    if any(mark.name in {'libxc', 'mgga'}
           for mark in item.iter_markers()):
        pytest.skip('No LibXC.')



import numpy as np
from functools import partial
from time import ctime

from ase.units import Hartree
from gpaw.utilities import convert_string_to_fd
from ase.utils.timing import Timer, timer

import gpaw
import gpaw.mpi as mpi
from gpaw.blacs import (BlacsGrid, BlacsDescriptor, Redistributor)
from gpaw.utilities.memory import maxrss
from gpaw.utilities.progressbar import ProgressBar
from gpaw.response.kspair import KohnShamPair, get_calc


class KohnShamLinearResponseFunction:
    r"""Class calculating linear response functions in the Kohn-Sham system

    Any linear response function can be calculated as a sum over transitions
    between the ground state and excited energy eigenstates.

    In the Kohn-Sham system this approach is particularly simple, as only
    excited states, for which a single electron has been moved from an occupied
    single-particle Kohn-Sham orbital to an unoccupied one, contribute.

    Resultantly, any linear response function in the Kohn-Sham system can be
    written as a sum over transitions between pairs of occupied and unoccupied
    Kohn-Sham orbitals.

    Currently, only collinear Kohn-Sham systems are supported. That is, all
    transitions can be written in terms of band indexes, k-points and spins:

    T (composit transition index): (n, k, s) -> (n', k', s')

    The sum over transitions is an integral over k-points in the 1st Brillouin
    Zone and a sum over all bands and spins. Sums over bands and spins can be
    handled together:

    t (composit transition index): (n, s) -> (n', s')

    __               __   __               __
    \      //        \    \      //        \
    /   =  ||dk dk'  /    /   =  ||dk dk'  /
    ‾‾     //        ‾‾   ‾‾     //        ‾‾
    T               n,n' s,s'              t
    """

    def __init__(self, gs, response=None, mode=None,
                 bandsummation='pairwise', nbands=None, kpointintegration=None,
                 world=mpi.world, nblocks=1, txt='-', timer=None):
        """Construct the KSLRF object

        Parameters
        ----------
        gs : str
            The groundstate calculation file that the linear response
            calculation is based on.
        response : str
            Type of response function.
            Currently, only susceptibilities are supported.
        mode: str
            Calculation mode.
            Currently, only a plane wave mode is implemented.
        bandsummation : str
            Band summation for pairs of Kohn-Sham orbitals
            'pairwise': sum over pairs of bands
            'double': double sum over band indices.
        nbands : int
            Maximum band index to include.
        kpointintegration : str
            Brillouin Zone integration for the Kohn-Sham orbital wave vector.
            Currently, only point integration is supported
        world : obj
            MPI communicator.
        nblocks : int
            Divide the response function storage into nblocks. Useful when the
            response function is large and memory requirements are restrictive.
        txt : str
            Output file.
        timer : func
            gpaw.utilities.timing.timer wrapper instance

        Attributes
        ----------
        kspair : gpaw.response.pair.KohnShamPair instance
            Class for handling pairs of Kohn-Sham orbitals
        pme : gpaw.response.pair.PairMatrixElement instance
            Class for calculating transition matrix elements for pairs of
            Kohn-Sham orbitals
        integrator : Integrator instance
            The integrator class is a general class for Brillouin Zone
            integration. The user defined integrand is integrated over k-points
            and summed over a given band and spin domain.

        Callables
        ---------
        self.add_integrand(kskptpair, weight, tmp_x, *args, **kwargs) : func
            Add the integrand for a given part of the domain to output array
        self.calculate(*args, **kwargs) : func
            Runs the calculation, returning the response function.
            Returned format can varry depending on response and mode.
        """
        # Output .txt filehandle
        self.fd = convert_string_to_fd(txt, world)
        self.cfd = self.fd
        print('Initializing KohnShamLinearResponseFunction', file=self.fd)

        # Communicators for parallelization
        self.world = world
        self.interblockcomm = None
        self.intrablockcomm = None
        self.initialize_communicators(nblocks)
        self.nblocks = self.interblockcomm.size

        # Timer
        self.timer = timer or Timer()

        # Load ground state calculation
        self.calc = get_calc(gs, fd=self.fd, timer=self.timer)

        # The KohnShamPair class handles data extraction from ground state
        self.kspair = KohnShamPair(self.calc, world=world,
                                   # Let each process handle slow steps only
                                   # for a fraction of all transitions.
                                   # t-transitions are distributed through
                                   # interblockcomm, k-points through
                                   # intrablockcomm.
                                   transitionblockscomm=self.interblockcomm,
                                   kptblockcomm=self.intrablockcomm,
                                   txt=self.fd, timer=self.timer)

        self.response = response
        self.mode = mode

        self.bandsummation = bandsummation
        self.nbands = nbands or self.calc.wfs.bd.nbands
        assert self.nbands <= self.calc.wfs.bd.nbands
        self.nocc1 = self.kspair.nocc1  # number of completely filled bands
        self.nocc2 = self.kspair.nocc2  # number of non-empty bands

        self.kpointintegration = kpointintegration
        self.integrator = create_integrator(self)
        # Each integrator might take some extra input kwargs
        self.extraintargs = {}

        # Attributes related to the specific response function
        self.pme = None

    def initialize_communicators(self, nblocks):
        """Set up MPI communicators to distribute the memory needed to store
        large arrays and parallelize calculations when possible.

        Parameters
        ----------
        nblocks : int
            Separate large arrays into n different blocks. Each process
            allocates memory for the large arrays. By allocating only a
            fraction/block of the total arrays, the memory requirements are
            eased.

        Sets
        ----
        interblockcomm : gpaw.mpi.Communicator
            Communicate between processes belonging to different memory blocks.
            In every communicator, there is one process for each block of
            memory, so that all blocks are represented.
            If nblocks < world.size, there will be size // nblocks different
            processes that allocate memory for the same block of the large
            arrays. Thus, there will be also size // nblocks different inter
            block communicators, grouping the processes into sets that allocate
            the entire arrays between them.
        intrablockcomm : gpaw.mpi.Communicator
            Communicate between processes belonging to the same memory block.
            There will be size // nblocks processes per memory block.
        """
        world = self.world
        if nblocks == 1:
            self.interblockcomm = world.new_communicator([world.rank])
            self.intrablockcomm = world
        else:
            assert world.size % nblocks == 0, world.size
            rank1 = world.rank // nblocks * nblocks
            rank2 = rank1 + nblocks
            self.interblockcomm = world.new_communicator(range(rank1, rank2))
            ranks = range(world.rank % nblocks, world.size, nblocks)
            self.intrablockcomm = world.new_communicator(ranks)
        print('Number of blocks:', nblocks, file=self.fd)

    @timer('Calculate Kohn-Sham linear response function')
    def calculate(self, spinrot=None, A_x=None):
        return self._calculate(spinrot, A_x)

    def _calculate(self, spinrot, A_x):
        """In-place calculation of the response function

        Parameters
        ----------
        spinrot : str
            Select spin rotation.
            Choices: 'u', 'd', '0' (= 'u' + 'd'), '-'= and '+'
            All rotations are included for spinrot=None ('0' + '+' + '-').
        A_x : ndarray
            Output array. If None, the output array is created.
        """
        self.spinrot = spinrot
        # Prepare to sum over bands and spins
        n1_t, n2_t, s1_t, s2_t = self.get_band_spin_transitions_domain()

        # Print information about the prepared calculation
        self.print_information(len(n1_t))
        if gpaw.dry_run:  # Exit after setting up
            print('    Dry run exit', file=self.fd)
            raise SystemExit

        A_x = self.setup_output_array(A_x)

        self.integrator.integrate(n1_t, n2_t, s1_t, s2_t,
                                  out_x=A_x, **self.extraintargs)

        # Different calculation modes might want the response function output
        # in different formats
        out = self.post_process(A_x)

        print('', file=self.cfd)

        return out

    def get_band_spin_transitions_domain(self):
        """Generate all allowed band and spin transitions.

        If only a subset of possible spin rotations are considered
        (examples: s1 = s2 or s2 = 1 - s1), do not include others
        in the sum over transitions.
        """
        n1_M, n2_M = get_band_transitions_domain(self.bandsummation,
                                                 self.nbands,
                                                 nocc1=self.nocc1,
                                                 nocc2=self.nocc2)
        s1_S, s2_S = get_spin_transitions_domain(self.bandsummation,
                                                 self.spinrot,
                                                 self.calc.wfs.nspins)

        return transitions_in_composite_index(n1_M, n2_M, s1_S, s2_S)

    def setup_output_array(self, A_x):
        raise NotImplementedError('Output array depends on mode')

    def get_ks_kpoint_pairs(self, k_pv, *args, **kwargs):
        raise NotImplementedError('Integrated pairs of states depend on'
                                  'response and mode')

    def initialize_pme(self, *args, **kwargs):
        raise NotImplementedError('Calculator method for matrix elements '
                                  'depend on response and mode')

    def calculate_pme(self, kskptpair, *args, **kwargs):
        raise NotImplementedError('Calculator method for matrix elements '
                                  'depend on response and mode')

    def add_integrand(self, kskptpair, weight, tmp_x, *args, **kwargs):
        raise NotImplementedError('Integrand depends on response and mode')

    def post_process(self, A_x):
        raise NotImplementedError('Post processing depends on mode')

    def print_information(self, nt):
        """Basic information about the input ground state, parallelization
        and sum over states"""
        ns = self.calc.wfs.nspins
        nbands = self.nbands
        nocc = self.nocc1
        npocc = self.nocc2
        nk = self.calc.wfs.kd.nbzkpts
        nik = self.calc.wfs.kd.nibzkpts

        if gpaw.dry_run:
            from gpaw.mpi import DryRunCommunicator
            size = gpaw.dry_run
            world = DryRunCommunicator(size)
        else:
            world = self.world
        wsize = world.size
        knsize = self.intrablockcomm.size
        bsize = self.interblockcomm.size

        spinrot = self.spinrot

        p = partial(print, file=self.cfd)

        p('Called a response.kslrf.KohnShamLinearResponseFunction.calculate()')
        p('%s' % ctime())
        p('Using a Kohn-Sham ground state with:')
        p('    Number of spins: %d' % ns)
        p('    Number of bands: %d' % nbands)
        p('    Number of completely occupied states: %d' % nocc)
        p('    Number of partially occupied states: %d' % npocc)
        p('    Number of kpoints: %d' % nk)
        p('    Number of irredicible kpoints: %d' % nik)
        p('')
        p('The response function calculation is performed in parallel with:')
        p('    world.size: %d' % wsize)
        p('    intrablockcomm.size: %d' % knsize)
        p('    interblockcomm.size: %d' % bsize)
        p('')
        p('The sum over band and spin transitions is performed using:')
        p('    Spin rotation: %s' % spinrot)
        p('    Total number of composite band and spin transitions: %d' % nt)
        p('')


def get_band_transitions_domain(bandsummation, nbands, nocc1=None, nocc2=None):
    """Get all pairs of bands to sum over

    Parameters
    ----------
    bandsummation : str
        Band summation method
    nbands : int
        number of bands
    nocc1 : int
        number of completely filled bands
    nocc2 : int
        number of non-empty bands

    Returns
    -------
    n1_M : ndarray
        band index 1, M = (n1, n2) composite index
    n2_M : ndarray
        band index 2, M = (n1, n2) composite index
    """
    _get_band_transitions_domain =\
        create_get_band_transitions_domain(bandsummation)
    n1_M, n2_M = _get_band_transitions_domain(nbands)

    return remove_null_transitions(n1_M, n2_M, nocc1=nocc1, nocc2=nocc2)


def create_get_band_transitions_domain(bandsummation):
    """Creator component deciding how to carry out band summation."""
    if bandsummation == 'pairwise':
        return get_pairwise_band_transitions_domain
    elif bandsummation == 'double':
        return get_double_band_transitions_domain
    raise ValueError(bandsummation)


def get_double_band_transitions_domain(nbands):
    """Make a simple double sum"""
    n_n = np.arange(0, nbands)
    m_m = np.arange(0, nbands)
    n_nm, m_nm = np.meshgrid(n_n, m_m)
    n_M, m_M = n_nm.flatten(), m_nm.flatten()

    return n_M, m_M


def get_pairwise_band_transitions_domain(nbands):
    """Make a sum over all pairs"""
    n_n = range(0, nbands)
    n_M = []
    m_M = []
    for n in n_n:
        m_m = range(n, nbands)
        n_M += [n] * len(m_m)
        m_M += m_m

    return np.array(n_M), np.array(m_M)


def remove_null_transitions(n1_M, n2_M, nocc1=None, nocc2=None):
    """Remove pairs of bands, between which transitions are impossible"""
    n1_newM = []
    n2_newM = []
    for n1, n2 in zip(n1_M, n2_M):
        if nocc1 is not None and (n1 < nocc1 and n2 < nocc1):
            continue  # both bands are fully occupied
        elif nocc2 is not None and (n1 >= nocc2 and n2 >= nocc2):
            continue  # both bands are completely unoccupied
        n1_newM.append(n1)
        n2_newM.append(n2)

    return np.array(n1_newM), np.array(n2_newM)


def get_spin_transitions_domain(bandsummation, spinrot, nspins):
    """Get structure of the sum over spins

    Parameters
    ----------
    bandsummation : str
        Band summation method
    spinrot : str
        spin rotation
    nspins : int
        number of spin channels in ground state calculation

    Returns
    -------
    s1_s : ndarray
        spin index 1, S = (s1, s2) composite index
    s2_S : ndarray
        spin index 2, S = (s1, s2) composite index
    """
    _get_spin_transitions_domain =\
        create_get_spin_transitions_domain(bandsummation)
    return _get_spin_transitions_domain(spinrot, nspins)


def create_get_spin_transitions_domain(bandsummation):
    """Creator component deciding how to carry out spin summation."""
    if bandsummation == 'pairwise':
        return get_pairwise_spin_transitions_domain
    elif bandsummation == 'double':
        return get_double_spin_transitions_domain
    raise ValueError(bandsummation)


def get_double_spin_transitions_domain(spinrot, nspins):
    """Usual spin rotations forward in time"""
    if nspins == 1:
        if spinrot is None or spinrot == '0':
            s1_S = [0]
            s2_S = [0]
        else:
            raise ValueError(spinrot, nspins)
    else:
        if spinrot is None:
            s1_S = [0, 0, 1, 1]
            s2_S = [0, 1, 0, 1]
        elif spinrot == '0':
            s1_S = [0, 1]
            s2_S = [0, 1]
        elif spinrot == 'u':
            s1_S = [0]
            s2_S = [0]
        elif spinrot == 'd':
            s1_S = [1]
            s2_S = [1]
        elif spinrot == '-':
            s1_S = [0]  # spin up
            s2_S = [1]  # spin down
        elif spinrot == '+':
            s1_S = [1]  # spin down
            s2_S = [0]  # spin up
        else:
            raise ValueError(spinrot)

    return np.array(s1_S), np.array(s2_S)


def get_pairwise_spin_transitions_domain(spinrot, nspins):
    """In a sum over pairs, transitions including a spin rotation may have to
    include terms, propagating backwards in time."""
    if spinrot in ['+', '-']:
        assert nspins == 2
        return np.array([0, 1]), np.array([1, 0])
    else:
        return get_double_spin_transitions_domain(spinrot, nspins)


def transitions_in_composite_index(n1_M, n2_M, s1_S, s2_S):
    """Use a composite index t for transitions (n, s) -> (n', s')."""
    n1_MS, s1_MS = np.meshgrid(n1_M, s1_S)
    n2_MS, s2_MS = np.meshgrid(n2_M, s2_S)
    return n1_MS.flatten(), n2_MS.flatten(), s1_MS.flatten(), s2_MS.flatten()


class PlaneWaveKSLRF(KohnShamLinearResponseFunction):
    """Class for doing KS-LRF calculations in plane wave mode"""

    def __init__(self, *args, eta=0.2, ecut=50, gammacentered=False,
                 disable_point_group=True, disable_time_reversal=True,
                 disable_non_symmorphic=True, bundle_integrals=True,
                 kpointintegration='point integration', bundle_kptpairs=False,
                 **kwargs):
        """Initialize the plane wave calculator mode.
        In plane wave mode, the linear response function is calculated for a
        given set of frequencies. The spatial part is expanded in plane waves
        for a given momentum transfer q within the first Brillouin Zone.

        Parameters
        ----------
        eta : float
            Energy broadening of spectra.
        ecut : float
            Energy cutoff for the plane wave representation.
        gammacentered : bool
            Center the grid of plane waves around the gamma point or q-vector.
        disable_point_group : bool
            Do not use the point group symmetry operators.
        disable_time_reversal : bool
            Do not use time reversal symmetry.
        disable_non_symmorphic : bool
            Do no use non symmorphic symmetry operators.
        bundle_integrals : bool
            Do the k-point integrals (large matrix multiplications)
            simultaneously for all frequencies.
            Can be switched of, if this step forces calculations out of memory.
        bundle_kptpairs : bool
            Extract the k-point pairs simultaneously, so no process has to wait
            for the others in the middle of the response function integration.
            [Only relevant in the case where ground state is distributed]
            Can be switched of, if this step forces calculations out of memory.
        """
        from gpaw.mpi import SerialCommunicator

        # Avoid any mode ambiguity
        if 'mode' in kwargs.keys():
            mode = kwargs.pop('mode')
            assert mode == 'pw'

        KSLRF = KohnShamLinearResponseFunction
        KSLRF.__init__(self, *args, mode='pw',
                       kpointintegration=kpointintegration, **kwargs)

        self.eta = eta / Hartree
        self.ecut = None if ecut is None else ecut / Hartree
        self.gammacentered = gammacentered

        self.disable_point_group = disable_point_group
        self.disable_time_reversal = disable_time_reversal
        self.disable_non_symmorphic = disable_non_symmorphic

        self.bundle_integrals = bundle_integrals

        # Bundle kptpairs if specified and using a distributed ground state
        cworld = self.calc.world
        if isinstance(cworld, SerialCommunicator) or cworld.size == 1:
            bundle_kptpairs = False
        self.bundle_kptpairs = bundle_kptpairs

        # Attributes related to specific q, given to self.calculate()
        self.pd = None  # Plane wave descriptor for given momentum transfer q
        self.pwsa = None  # Plane wave symmetry analyzer for given q
        self.wd = None  # Frequency descriptor for the given frequencies
        self.omega_w = None  # Frequencies in code units

    @timer('Calculate Kohn-Sham linear response function in plane wave mode')
    def calculate(self, q_c, frequencies, spinrot=None, A_x=None):
        """
        Parameters
        ----------
        q_c : list or ndarray or PWDescriptor
            Momentum transfer (and possibly plane wave basis)
        frequencies : ndarray or FrequencyDescriptor
            Array of frequencies to evaluate the response function at or
            descriptor of those frequencies.

        Returns
        -------
        pd : Planewave descriptor
            Planewave descriptor for q_c.
        A_wGG : ndarray
            The linear response function.
        """
        # Set up plane wave description with the gived momentum transfer q
        self.pd = self.get_PWDescriptor(q_c)
        self.pwsa = self.get_PWSymmetryAnalyzer(self.pd)

        # Set up frequency descriptor for the given frequencies
        self.wd = self.get_FreqDescriptor(frequencies)
        self.omega_w = self.wd.get_data()

        # In-place calculation
        return self._calculate(spinrot, A_x)

    def get_PWDescriptor(self, q_c):
        """Get the planewave descriptor for a certain momentum transfer q_c."""
        from gpaw.pw.descriptor import PWDescriptor
        if isinstance(q_c, PWDescriptor):
            return q_c
        else:
            from gpaw.kpt_descriptor import KPointDescriptor
            q_c = np.asarray(q_c, dtype=float)
            qd = KPointDescriptor([q_c])
            pd = PWDescriptor(self.ecut, self.calc.wfs.gd,
                              complex, qd, gammacentered=self.gammacentered)
            return pd

    @timer('Get PW symmetry analyser')
    def get_PWSymmetryAnalyzer(self, pd):
        from gpaw.response.pair import PWSymmetryAnalyzer as PWSA

        return PWSA(self.calc.wfs.kd, pd,
                    timer=self.timer, txt=self.fd,
                    disable_point_group=self.disable_point_group,
                    disable_time_reversal=self.disable_time_reversal,
                    disable_non_symmorphic=self.disable_non_symmorphic)

    def get_FreqDescriptor(self, frequencies):
        """Get the frequency descriptor for a certain input frequencies."""
        if isinstance(frequencies, FrequencyDescriptor):
            return frequencies
        else:
            return FrequencyDescriptor(np.asarray(frequencies) / Hartree)

    def print_information(self, nt):
        """Basic information about the input ground state, parallelization,
        sum over states and calculated response function array."""
        KohnShamLinearResponseFunction.print_information(self, nt)

        pd = self.pd
        q_c = pd.kd.bzk_kc[0]
        nw = len(self.omega_w)
        eta = self.eta * Hartree
        ecut = self.ecut * Hartree
        ngmax = pd.ngmax
        Asize = nw * pd.ngmax**2 * 16. / 1024**2 / self.interblockcomm.size

        p = partial(print, file=self.cfd)

        p('The response function is calculated in the PlaneWave mode, using:')
        p('    q_c: [%f, %f, %f]' % (q_c[0], q_c[1], q_c[2]))
        p('    Number of frequency points: %d' % nw)
        p('    Broadening (eta): %f' % eta)
        p('    Planewave cutoff: %f' % ecut)
        p('    Number of planewaves: %d' % ngmax)
        p('')
        p('    Memory estimates:')
        p('        A_wGG: %f M / cpu' % Asize)
        p('        Memory usage before allocation: %f M / cpu' % (maxrss() /
                                                                  1024**2))
        p('')

    def setup_output_array(self, A_x=None):
        """Initialize the output array in blocks"""
        # Could use some more documentation XXX
        nG = self.pd.ngmax
        nw = len(self.omega_w)
        mynG = (nG + self.interblockcomm.size - 1) // self.interblockcomm.size
        self.Ga = min(self.interblockcomm.rank * mynG, nG)
        self.Gb = min(self.Ga + mynG, nG)
        # if self.interblockcomm.rank == 0:
        #     assert self.Gb - self.Ga >= 3
        # assert mynG * (self.interblockcomm.size - 1) < nG
        if self.bundle_integrals:
            # Setup A_GwG
            if A_x is not None:
                nx = nw * (self.Gb - self.Ga) * nG
                A_GwG = A_x[:nx].reshape((nG, nw, self.Gb - self.Ga))
                A_GwG[:] = 0.0
            else:
                A_GwG = np.zeros((nG, nw, self.Gb - self.Ga), complex)

            return A_GwG
        else:
            # Setup A_wGG
            if A_x is not None:
                nx = nw * (self.Gb - self.Ga) * nG
                A_wGG = A_x[:nx].reshape((nw, self.Gb - self.Ga, nG))
                A_wGG[:] = 0.0
            else:
                A_wGG = np.zeros((nw, self.Gb - self.Ga, nG), complex)

            return A_wGG

    def get_ks_kpoint_pairs(self, k_pv, n1_t, n2_t, s1_t, s2_t):
        """Get all pairs of Kohn-Sham transitions:

        (n1_t, k_c, s1_t) -> (n2_t, k_c + q_c, s2_t)

        for each process with its own k-point.
        """
        k_pc = np.array([np.dot(self.pd.gd.cell_cv, k_v) / (2 * np.pi)
                         for k_v in k_pv])
        q_c = self.pd.kd.bzk_kc[0]
        return self.kspair.get_kpoint_pairs(n1_t, n2_t, k_pc, k_pc + q_c,
                                            s1_t, s2_t)

    def initialize_pme(self):
        self.pme.initialize(self.pd)

    def calculate_pme(self, kskptpair):
        self.pme(kskptpair, self.pd)

    def add_integrand(self, kskptpair, weight, tmp_x, **kwargs):
        raise NotImplementedError('Integrand depends on response')

    @timer('Post processing')
    def post_process(self, A_x):
        if self.bundle_integrals:
            # A_x = A_GwG
            A_wGG = A_x.transpose((1, 2, 0))
        else:
            A_wGG = A_x

        tmpA_wGG = self.redistribute(A_wGG)  # distribute over frequencies
        with self.timer('Symmetrizing Kohn-Sham linear response function'):
            self.pwsa.symmetrize_wGG(tmpA_wGG)
        self.redistribute(tmpA_wGG, A_wGG)

        return self.pd, A_wGG

    @timer('Redistribute memory')
    def redistribute(self, in_wGG, out_x=None):
        """Redistribute array.

        Switch between two kinds of parallel distributions:

        1) parallel over G-vectors (second dimension of in_wGG)
        2) parallel over frequency (first dimension of in_wGG)

        Returns new array using the memory in the 1-d array out_x.
        """

        comm = self.interblockcomm

        if comm.size == 1:
            return in_wGG

        nw = len(self.omega_w)
        nG = in_wGG.shape[2]
        mynw = (nw + comm.size - 1) // comm.size
        mynG = (nG + comm.size - 1) // comm.size

        bg1 = BlacsGrid(comm, comm.size, 1)
        bg2 = BlacsGrid(comm, 1, comm.size)
        md1 = BlacsDescriptor(bg1, nw, nG**2, mynw, nG**2)
        md2 = BlacsDescriptor(bg2, nw, nG**2, nw, mynG * nG)

        if len(in_wGG) == nw:
            mdin = md2
            mdout = md1
        else:
            mdin = md1
            mdout = md2

        r = Redistributor(comm, mdin, mdout)

        outshape = (mdout.shape[0], mdout.shape[1] // nG, nG)
        if out_x is None:
            out_wGG = np.empty(outshape, complex)
        else:
            out_wGG = out_x[:np.product(outshape)].reshape(outshape)

        r.redistribute(in_wGG.reshape(mdin.shape),
                       out_wGG.reshape(mdout.shape))

        return out_wGG


class FrequencyDescriptor:
    """Describes a one-dimensional array of frequencies."""

    def __init__(self, data_x):
        self.data_x = np.array(np.sort(data_x))
        self._data_len = len(data_x)

    def __len__(self):
        return self._data_len

    def get_data(self):
        return self.data_x


class Integrator:
    """Baseclass for integrating over k-points in the first Brillouin Zone
    and summing over bands and spin.
    """
    def __init__(self, kslrf):
        """
        Parameters
        ----------
        kslrf : KohnShamLinearResponseFunction instance
        """
        self.kslrf = kslrf
        self.timer = self.kslrf.timer

    def slice_kpoint_domain(self, bzk_kv, weight_k):
        """When integrating over k-points, slice the domain in pieces with one
        k-point per process each.

        Returns
        -------
        bzk_ipv : nd.array
            k-points coordinates for each process for each iteration
        """
        nk = bzk_kv.shape[0]
        size = self.kslrf.intrablockcomm.size
        ni = (nk + size - 1) // size
        bzk_ipv = np.array([bzk_kv[i * size:(i + 1) * size]
                            for i in range(ni)])

        # Extract the weight corresponding to the process' own k-point pair
        weight_ip = np.array([weight_k[i * size:(i + 1) * size]
                              for i in range(ni)])
        weight_i = [None] * len(weight_ip)
        krank = self.kslrf.intrablockcomm.rank
        for i, w_p in enumerate(weight_ip):
            if krank in range(len(w_p)):
                weight_i[i] = w_p[krank]

        return bzk_ipv, weight_i

    @timer('Integrate response function')
    def integrate(self, n1_t, n2_t, s1_t, s2_t,
                  out_x=None, **kwargs):
        if out_x is None:
            raise NotImplementedError

        bzk_kv, weight_k = self.get_kpoint_domain()
        prefactor = self.calculate_bzint_prefactor(bzk_kv)
        out_x /= prefactor
        self._integrate(bzk_kv, weight_k,
                        n1_t, n2_t, s1_t, s2_t, out_x, **kwargs)
        out_x *= prefactor

        return out_x

    def get_kpoint_domain(self):
        raise NotImplementedError('Domain depends on integration method')

    def calculate_bzint_prefactor(self, bzk_kv):
        raise NotImplementedError('Prefactor depends on integration method')

    def _integrate(self, bzk_kv, weight_k,
                   n1_t, n2_t, s1_t, s2_t, out_x, **kwargs):
        raise NotImplementedError('Integration method is defined by subclass')


class PWPointIntegrator(Integrator):
    """A simple point integrator for the plane wave mode."""

    @timer('Get k-point domain')
    def get_kpoint_domain(self):
        # Could use some more documentation XXX
        K_gK = self.kslrf.pwsa.group_kpoints()
        bzk_kc = np.array([self.kslrf.calc.wfs.kd.bzk_kc[K_K[0]] for
                           K_K in K_gK])

        bzk_kv = np.dot(bzk_kc, self.kslrf.pd.gd.icell_cv) * 2 * np.pi

        nsym = self.kslrf.pwsa.how_many_symmetries()
        weight_k = [self.kslrf.pwsa.get_kpoint_weight(k_c) / nsym
                    for k_c in bzk_kc]

        return bzk_kv, weight_k

    def calculate_bzint_prefactor(self, bzk_kv):
        # Could use some more documentation XXX
        if self.kslrf.calc.wfs.kd.refine_info is not None:
            nbzkpts = self.kslrf.calc.wfs.kd.refine_info.mhnbzkpts
        else:
            nbzkpts = self.kslrf.calc.wfs.kd.nbzkpts
        frac = len(bzk_kv) / nbzkpts

        return (2 * frac * self.kslrf.pwsa.how_many_symmetries() /
                (self.kslrf.calc.wfs.nspins * (2 * np.pi)**3))

    def _integrate(self, bzk_kv, weight_k,
                   n1_t, n2_t, s1_t, s2_t, out_x, **kwargs):
        """Do a simple sum over k-points in the first Brillouin Zone,
        adding the integrand (summed over bands and spin) for each k-point."""

        nk = bzk_kv.shape[0]
        vol = abs(np.linalg.det(self.kslrf.calc.wfs.gd.cell_cv))

        kpointvol = (2 * np.pi)**3 / vol / nk
        out_x /= kpointvol

        # Initialize pme
        print('----------',
              file=self.kslrf.cfd)
        print('Initializing PairMatrixElement',
              file=self.kslrf.cfd, flush=True)
        self.kslrf.initialize_pme()

        # Slice domain
        bzk_ipv, weight_i = self.slice_kpoint_domain(bzk_kv, weight_k)

        # Perform sum over k-points
        tmp_x = np.zeros_like(out_x)
        pb = ProgressBar(self.kslrf.cfd)
        if self.kslrf.bundle_kptpairs:
            # Extract all the process' kptpairs at once, then do integration
            print('\nExtracting Kohn-Sham k-point pairs',
                  file=self.kslrf.cfd, flush=True)
            kskptpair_i = []
            for k_pv in bzk_ipv:
                kskptpair_i.append(self.kslrf.get_ks_kpoint_pairs(k_pv,
                                                                  n1_t, n2_t,
                                                                  s1_t, s2_t))

            print('\nIntegrating response function',
                  file=self.kslrf.cfd, flush=True)
            # Carry out sum
            for i, kskptpair in pb.enumerate(kskptpair_i):
                if kskptpair is not None:
                    weight = weight_i[i]
                    assert weight is not None
                    self.kslrf.calculate_pme(kskptpair)
                    self.kslrf.add_integrand(kskptpair, weight,
                                             tmp_x, **kwargs)
        else:
            # Each process will do its own k-points, but it has to follow the
            # others, as it may have to send them information about its
            # partition of the ground state
            print('\nIntegrating response function',
                  file=self.kslrf.cfd, flush=True)
            for i, k_pv in pb.enumerate(bzk_ipv):
                kskptpair = self.kslrf.get_ks_kpoint_pairs(k_pv, n1_t, n2_t,
                                                           s1_t, s2_t)
                if kskptpair is not None:
                    weight = weight_i[i]
                    assert weight is not None
                    self.kslrf.calculate_pme(kskptpair)
                    self.kslrf.add_integrand(kskptpair, weight,
                                             tmp_x, **kwargs)

        # Sum over the k-points that have been distributed between processes
        with self.timer('Sum over distributed k-points'):
            self.kslrf.intrablockcomm.sum(tmp_x)

        out_x += tmp_x
        out_x *= kpointvol


def create_integrator(kslrf):
    """Creator component for the integrator"""
    if kslrf.mode == 'pw':
        if kslrf.kpointintegration is None or \
           kslrf.kpointintegration == 'point integration':
            return PWPointIntegrator(kslrf)

    raise ValueError(kslrf.mode, kslrf.kpointintegration)



from __future__ import annotations
from typing import TYPE_CHECKING
import numpy as np
from ase.dft.kpoints import monkhorst_pack
from ase.utils import plural
from gpaw.mpi import MPIComm
from gpaw.typing import Array1D
if TYPE_CHECKING:
    from gpaw.new.symmetry import Symmetries


class BZPoints:
    def __init__(self, points):
        self.kpt_Kc = points
        self.gamma_only = len(self.kpt_Kc) == 1 and not self.kpt_Kc.any()

    def __len__(self):
        """Number of k-points in the IBZ."""
        return len(self.kpt_Kc)

    def __repr__(self):
        return f'BZPoints([<{len(self)} points>])'

    def __str__(self):
        if self.gamma_only:
            return '1 k-point (Gamma)'
        return f'{len(self)} k-points'


class MonkhorstPackKPoints(BZPoints):
    def __init__(self, size, shift=(0, 0, 0)):
        self.size_c = size
        self.shift_c = np.array(shift)
        super().__init__(monkhorst_pack(size) + shift)

    def __repr__(self):
        return f'MonkhorstPackKPoints({self.size_c}, shift={self.shift_c})'

    def __str__(self):
        if self.gamma_only:
            return '1 k-point (Gamma)'

        a, b, c = self.size_c
        s = f'{len(self)} k-points: {a} x {b} x {c} Monkhorst-Pack grid'

        if self.shift_c.any():
            s += ' + ['
            for x in self.shift_c:
                if x != 0 and abs(round(1 / x) - 1 / x) < 1e-12:
                    s += '1/%d,' % round(1 / x)
                else:
                    s += f'{x:f},'
            s = s[:-1] + ']'

        return s


class IBZ:
    def __init__(self,
                 symmetries: Symmetries,
                 bz: BZPoints,
                 ibz2bz, bz2ibz, weights):
        self.symmetries = symmetries
        self.bz = bz
        self.weight_k = weights
        self.kpt_kc = bz.kpt_Kc[ibz2bz]
        self.ibz2bz_k = ibz2bz
        self.bz2ibz_K = bz2ibz

        # self.bz2bz_Ks = []  # later ...

    def __len__(self):
        """Number of k-points in the IBZ."""
        return len(self.kpt_kc)

    def __repr__(self):
        return f'IBZ(<{plural(len(self), "point")}>)'

    def __str__(self):
        s = ''
        # if -1 in self.bz2bz_Ks:
        #    s += 'Note: your k-points are not as symmetric as your crystal!\n'
        N = len(self)
        s += str(self.bz)
        nk = plural(N, 'k-point')
        s += f'\n{nk} in the irreducible part of the Brillouin zone\n'

        if isinstance(self.bz, MonkhorstPackKPoints):
            w_k = (self.weight_k * len(self.bz)).round().astype(int)

        s += '          k-point in crystal coordinates           weight\n'
        for k, (a, b, c) in enumerate(self.kpt_kc):
            if k >= 10 and k < N - 1:
                continue
            elif k == 10:
                s += '          ...\n'
            s += f'{k:4}:   {a:12.8f}  {b:12.8f}  {c:12.8f}     '
            if isinstance(self.bz, MonkhorstPackKPoints):
                s += f'{w_k[k]}/{len(self.bz)}\n'
            else:
                s += f'{self.weight_k[k]:.8f}\n'
        return s

    def ranks(self, comm: MPIComm) -> Array1D:
        """Distribute k-points over MPI-communicator."""
        return ranks(comm.size, len(self))


def ranks(N, K) -> Array1D:
    """Distribute k-points over MPI-communicator.

    >>> ranks(4, 6)
    array([0, 1, 2, 2, 3, 3])
    """
    n, x = divmod(K, N)

    rnks = np.empty(K, int)
    k0 = K - x * (n + 1)
    for k in range(k0):
        rnks[k] = k // n
    for k in range(k0, K):
        rnks[k] = (k - k0) // (n + 1) + x
    return rnks



from collections import defaultdict
from contextlib import contextmanager
from functools import lru_cache
from time import time


def cached_property(method):
    """Quick'n'dirty implementation of cached_property coming in Python 3.8."""
    return property(lru_cache(maxsize=None)(method))


class Timer:
    def __init__(self):
        self.times = defaultdict(float)
        self.times['Total'] = -time()

    @contextmanager
    def __call__(self, name):
        t1 = time()
        try:
            yield
        finally:
            t2 = time()
            self.times[name] += t2 - t1

    def write(self, log):
        self.times['Total'] += time()
        total = self.times['Total']
        log()
        for name, t in self.times.items():
            n = int(round(40 * t / total))
            if n == 0:
                bar = '|'
            else:
                bar = '|' + (n - 1) * '-' + '|'
            log(f'Time ({name + "):":12}{t:10.3f} seconds', bar)



import numpy as np

from gpaw.mpi import MPIComm
from gpaw.new.brillouin import IBZ, BZPoints
from gpaw.rotation import rotation


class Symmetries:
    def __init__(self, symmetry):
        self.symmetry = symmetry
        self.rotation_scc = symmetry.op_scc
        self.translation_sc = symmetry.ft_sc
        self.a_sa = symmetry.a_sa
        cell_cv = symmetry.cell_cv
        self.rotation_svv = np.einsum('vc, scd, dw -> svw',
                                      np.linalg.inv(cell_cv),
                                      self.rotation_scc,
                                      cell_cv)
        self.rotation_lsmm = [
            np.array([rotation(l, r_vv) for r_vv in self.rotation_svv])
            for l in range(4)]
        self._rotations = {}

    def __len__(self):
        return len(self.rotation_scc)

    def __str__(self):
        return str(self.symmetry)

    def reduce(self,
               bz: BZPoints,
               comm: MPIComm = None) -> IBZ:
        (_, weight_k, sym_k, time_reversal_k, bz2ibz_K, ibz2bz_k,
         bz2bz_Ks) = self.symmetry.reduce(bz.kpt_Kc, comm)

        return IBZ(self, bz, ibz2bz_k, bz2ibz_K, weight_k)

    def check_positions(self, fracpos_ac):
        self.symmetry.check(fracpos_ac)

    def symmetrize_forces(self, F_av):
        return self.symmetry.symmetrize_forces(F_av)

    def rotations(self, l_j):
        ells = tuple(l_j)
        rotation_sii = self._rotations.get(ells)
        if rotation_sii is None:
            ni = sum(2 * l + 1 for l in l_j)
            rotation_sii = np.zeros((len(self), ni, ni))
            i1 = 0
            for l in l_j:
                i2 = i1 + 2 * l + 1
                rotation_sii[:, i1:i2, i1:i2] = self.rotation_lsmm[l]
                i1 = i2
            self._rotations[ells] = rotation_sii
        return rotation_sii



from gpaw.mpi import MPIComm, world
import sys
import os
from pathlib import Path
from gpaw.utilities.memory import maxrss


class Logger:
    def __init__(self,
                 filename='-',
                 comm: MPIComm = None):
        comm = comm or world

        if comm.rank > 0 or filename is None:
            self.fd = open(os.devnull, 'w')
            self.close_fd = True
        elif filename == '-':
            self.fd = sys.stdout
            self.close_fd = False
        elif isinstance(filename, (str, Path)):
            self.fd = open(filename, 'w')
            self.close_fd = True
        else:
            self.fd = filename
            self.close_fd = False

    def __del__(self) -> None:
        try:
            mib = maxrss() / 1024**2
        except (NameError, LookupError):
            pass
        else:
            self.fd.write(f'\nMax RSS: {mib:.3f} MiB\n')
        if self.close_fd:
            self.fd.close()

    def __call__(self, *args, **kwargs) -> None:
        print(*args, **kwargs, file=self.fd)



from typing import Any, TYPE_CHECKING, Union, Tuple
import numpy as np

if TYPE_CHECKING:
    from numpy.typing import ArrayLike
else:
    ArrayLike = Any  # type: ignore

ArrayLike1D = ArrayLike
ArrayLike2D = ArrayLike

ArrayND = np.ndarray
Array1D = ArrayND
Array2D = ArrayND
Array3D = ArrayND
Array4D = ArrayND

Vector = Union[Tuple[float, float, float], Array1D]



from __future__ import annotations

from pathlib import Path
from typing import IO, Any, Union

from ase import Atoms
from ase.units import Bohr, Ha

from gpaw import __version__
from gpaw.new import Timer
from gpaw.new.calculation import DFTCalculation
from gpaw.new.input_parameters import InputParameters
from gpaw.new.logger import Logger
from gpaw.new.old import OldStuff, read_gpw
from gpaw.typing import Array1D, Array2D


def GPAW(filename: Union[str, Path, IO[str]] = None,
         **kwargs) -> ASECalculator:
    """Create ASE-compatible GPAW calculator."""
    params = InputParameters(kwargs)
    txt = params.txt
    if txt == '?':
        txt = '-' if filename is None else None
    world = params.parallel['world']
    log = Logger(txt, world)

    if filename is not None:
        kwargs.pop('txt', None)
        assert len(kwargs) == 0
        calculation, params = read_gpw(filename, log, params.parallel)
        return ASECalculator(params, log, calculation)

    write_header(log, world, kwargs)
    return ASECalculator(params, log)


class ASECalculator(OldStuff):
    """This is the ASE-calculator frontend for doing a GPAW calculation."""
    def __init__(self,
                 params: InputParameters,
                 log: Logger,
                 calculation=None):
        self.params = params
        self.log = log
        self.calculation = calculation

        self.atoms = None
        self.timer = Timer()

    def calculate_property(self, atoms: Atoms, prop: str) -> Any:
        """Calculate (if not already calculated) a property.

        Must be one of

        * energy
        * forces
        * stress
        * magmom
        * magmoms
        * dipole
        """
        log = self.log

        if self.calculation is not None:
            changes = compare_atoms(self.atoms, atoms)
            if changes & {'numbers', 'pbc', 'cell'}:
                if 'numbers' not in changes:
                    magmom_a = self.calculation.results.get('magmoms')
                    if magmom_a is not None:
                        atoms = atoms.copy()
                        atoms.set_initial_magnetic_moments(magmom_a)
                self.calculation = None

        if self.calculation is None:
            self.calculation = self.create_new_calculation(atoms)
            self.converge(atoms)
        elif changes:
            self.move_atoms(atoms)
            self.converge(atoms)

        if prop not in self.calculation.results:
            if prop == 'forces':
                with self.timer('Forces'):
                    self.calculation.forces(log)
            elif prop == 'stress':
                with self.timer('Stress'):
                    self.calculation.stress(log)
            else:
                raise ValueError('Unknown property:', prop)

        return self.calculation.results[prop]

    def create_new_calculation(self, atoms: Atoms) -> DFTCalculation:
        with self.timer('init'):
            calculation = DFTCalculation.from_parameters(atoms, self.params,
                                                         self.log)
        return calculation

    def move_atoms(self, atoms):
        with self.timer('move'):
            self.calculation = self.calculation.move_atoms(atoms, self.log)

    def converge(self, atoms):
        """Iterate to self-consistent solution.

        Will also calculate "cheap" properties: energy, magnetic moments
        and dipole moment.
        """
        with self.timer('SCF'):
            self.calculation.converge(self.log)
        self.calculation.energies(self.log)
        # self.calculation.dipole(self.log)
        self.calculation.magmoms(self.log)
        self.atoms = atoms.copy()
        self.calculation.write_converged(self.log)

    def __del__(self):
        self.timer.write(self.log)

    def get_potential_energy(self,
                             atoms: Atoms,
                             force_consistent: bool = False) -> float:
        return self.calculate_property(atoms,
                                       'free_energy' if force_consistent else
                                       'energy') * Ha

    def get_forces(self, atoms: Atoms) -> Array2D:
        return self.calculate_property(atoms, 'forces') * (Ha / Bohr)

    def get_stress(self, atoms: Atoms) -> Array1D:
        return self.calculate_property(atoms, 'stress') * (Ha / Bohr**3)


def write_header(log, world, kwargs):
    from gpaw.io.logger import write_header as header
    log(f' __  _  _\n| _ |_)|_||  |\n|__||  | ||/\\| - {__version__}\n')
    header(log, world)
    log('Input parameters = {\n    ', end='')
    log(',\n    '.join(f'{k!r}: {v!r}' for k, v in kwargs.items()) + '}')


def compare_atoms(a1: Atoms, a2: Atoms) -> set[str]:
    if len(a1.numbers) != len(a2.numbers) or (a1.numbers != a2.numbers).any():
        return {'numbers'}
    if (a1.pbc != a2.pbc).any():
        return {'pbc'}
    if abs(a1.cell - a2.cell).max() > 0.0:
        return {'cell'}
    if abs(a1.positions - a2.positions).max() > 0.0:
        return {'positions'}
    return set()
